#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.lib.services.network import base


class TapServicesClient(base.BaseNetworkClient):

    def create_tap_service(self, **kwargs):
        uri = '/taas/tap_services'
        post_data = {'tap_service': kwargs}
        return self.create_resource(uri, post_data)

    def update_tap_service(self, tap_service_id, **kwargs):
        uri = '/taas/tap_services/%s' % tap_service_id
        post_data = {'tap_service': kwargs}
        return self.update_resource(uri, post_data)

    def show_tap_service(self, tap_service_id, **fields):
        uri = '/taas/tap_services/%s' % tap_service_id
        return self.show_resource(uri, **fields)

    def delete_tap_service(self, tap_service_id):
        uri = '/taas/tap_services/%s' % tap_service_id
        return self.delete_resource(uri)

    def list_tap_services(self, **filters):
        uri = '/taas/tap_services'
        return self.list_resources(uri, **filters)


class TapFlowsClient(base.BaseNetworkClient):

    def create_tap_flow(self, **kwargs):
        uri = '/taas/tap_flows'
        post_data = {'tap_flow': kwargs}
        return self.create_resource(uri, post_data)

    def update_tap_flow(self, tap_flow_id, **kwargs):
        uri = '/taas/tap_flows/%s' % tap_flow_id
        post_data = {'tap_flow': kwargs}
        return self.update_resource(uri, post_data)

    def show_tap_flow(self, tap_flow_id, **fields):
        uri = '/taas/tap_flows/%s' % tap_flow_id
        return self.show_resource(uri, **fields)

    def delete_tap_flow(self, tap_flow_id):
        uri = '/taas/tap_flows/%s' % tap_flow_id
        return self.delete_resource(uri)

    def list_tap_flows(self, **filters):
        uri = '/taas/tap_flows'
        return self.list_resources(uri, **filters)


class TapMirrorsClient(base.BaseNetworkClient):
    def create_tap_mirror(self, **kwargs):
        uri = '/taas/tap_mirrors'
        post_data = {'tap_mirror': kwargs}
        return self.create_resource(uri, post_data)

    def update_tap_mirror(self, tap_mirror_id, **kwargs):
        uri = '/taas/tap_mirrors/%s' % tap_mirror_id
        post_data = {'tap_mirror': kwargs}
        return self.update_resource(uri, post_data)

    def show_tap_mirror(self, tap_mirror_id, **fields):
        uri = '/taas/tap_mirrors/%s' % tap_mirror_id
        return self.show_resource(uri, **fields)

    def delete_tap_mirror(self, tap_mirror_id):
        uri = '/taas/tap_mirrors/%s' % tap_mirror_id
        return self.delete_resource(uri)

    def list_tap_mirrors(self, **filters):
        uri = '/taas/tap_mirrors'
        return self.list_resources(uri, **filters)



# Copyright 2017 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from tempest.common import utils
from tempest.common import waiters
from tempest.lib import decorators

from neutron_tempest_plugin.common import ssh
from neutron_tempest_plugin import config
from neutron_tempest_plugin.scenario import base
from neutron_tempest_plugin.scenario import constants as const

CONF = config.CONF


class FloatingIpTestCasesAdmin(base.BaseTempestTestCase):
    credentials = ['primary', 'admin']

    @classmethod
    def setup_clients(cls):
        super().setup_clients()
        # admin_client set in BaseAdminNetworkTest but here we inherit from
        # BaseNetworkTest
        if not cls.admin_client:
            cls.admin_client = cls.os_admin.network_client

    @classmethod
    @utils.requires_ext(extension="router", service="network")
    def resource_setup(cls):
        super().resource_setup()
        cls.network = cls.create_network()
        cls.create_subnet(cls.network)
        router = cls.create_router_by_client()
        cls.create_router_interface(router['id'], cls.subnets[0]['id'])
        # Create keypair with admin privileges
        cls.keypair = cls.create_keypair(client=cls.os_admin.keypairs_client)

        # Create security group with admin privileges
        network_client = cls.os_admin.network_client
        cls.secgroup = cls.create_security_group(
            client=cls.os_admin.network_client)
        cls.create_loginable_secgroup_rule(
            secgroup_id=cls.secgroup['id'],
            client=network_client)
        cls.create_pingable_secgroup_rule(
            secgroup_id=cls.secgroup['id'],
            client=network_client),

    def _choose_az_and_node(self):
        az_list = self.os_admin.az_client.list_availability_zones(
            detail=True)['availabilityZoneInfo']
        hv_list = self.os_admin.hv_client.list_hypervisors()['hypervisors']
        for az in az_list:
            if not az['zoneState']['available']:
                continue
            for host, services in az['hosts'].items():
                for service, info in services.items():
                    if (
                        service == 'nova-compute' and
                        info['active'] and info['available']
                    ):
                        hv = [
                            h for h in hv_list
                            if (
                                h['hypervisor_hostname'].startswith(host) and
                                h["state"] == "up" and
                                h["status"] == "enabled"
                            )
                        ]
                        if not hv:
                            continue
                        return az['zoneName'], hv[0]['hypervisor_hostname']
        return None, None

    def _create_vms(self, hyper, avail_zone, num_servers=2):
        servers, fips, server_ssh_clients = ([], [], [])
        # Create the availability zone with default zone and
        # a specific mentioned hypervisor.
        az = avail_zone + ':' + hyper
        for i in range(num_servers):
            servers.append(self.create_server(
                flavor_ref=CONF.compute.flavor_ref,
                image_ref=CONF.compute.image_ref,
                key_name=self.keypair['name'],
                networks=[{'uuid': self.network['id']}],
                security_groups=[{'name': self.secgroup['name']}],
                availability_zone=az))
        for i, server in enumerate(servers):
            waiters.wait_for_server_status(
                self.os_admin.servers_client, server['server']['id'],
                const.SERVER_STATUS_ACTIVE)
            port = self.admin_client.list_ports(
                network_id=self.network['id'],
                device_id=server['server']['id']
            )['ports'][0]
            fip = self.create_floatingip(port=port,
                                         client=self.os_admin.network_client)
            fips.append(fip)
            server_ssh_clients.append(ssh.Client(
                fips[i]['floating_ip_address'], CONF.validation.image_ssh_user,
                pkey=self.keypair['private_key']))
        return servers, server_ssh_clients, fips

    @decorators.idempotent_id('6bba729b-3fb6-494b-9e1e-82bbd89a1045')
    def test_two_vms_fips(self):
        """Test two VMs floating IPs

        This test verifies the ability of two instances
        that were created in the same compute node and same availability zone
        to reach each other.
        """
        avail_zone, hyper = self._choose_az_and_node()
        if not (avail_zone and hyper):
            self.fail("No compute host is available")
        servers, server_ssh_clients, fips = self._create_vms(hyper, avail_zone)
        self.check_remote_connectivity(
            server_ssh_clients[0], fips[1]['floating_ip_address'],
            servers=servers)



# Copyright (c) 2015 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.lib.common.utils import data_utils
from tempest.lib import decorators
from tempest.lib import exceptions as lib_exc

from neutron_tempest_plugin.api import test_address_scopes


class AddressScopeTestNegative(test_address_scopes.AddressScopeTestBase):

    @decorators.attr(type='negative')
    @decorators.idempotent_id('9c92ec34-0c50-4104-aa47-9ce98d5088df')
    def test_tenant_create_shared_address_scope(self):
        self.assertRaises(lib_exc.Forbidden, self._create_address_scope,
                          shared=True, ip_version=4)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('a857b61e-bf53-4fab-b21a-b0daaf81b5bd')
    def test_tenant_update_address_scope_shared_true(self):
        self.assertRaises(lib_exc.Forbidden,
                          self._test_update_address_scope_helper, shared=True)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('a859ef2f-9c76-4e2e-ba0f-e0339a489e8c')
    def test_tenant_update_address_scope_shared_false(self):
        self.assertRaises(lib_exc.Forbidden,
                          self._test_update_address_scope_helper, shared=False)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('9b6dd7ad-cabb-4f55-bd5e-e61176ef41f6')
    def test_get_non_existent_address_scope(self):
        non_exist_id = data_utils.rand_name('address_scope')
        self.assertRaises(lib_exc.NotFound, self.client.show_address_scope,
                          non_exist_id)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('ef213552-f2da-487d-bf4a-e1705d115ff1')
    def test_tenant_get_not_shared_admin_address_scope(self):
        address_scope = self._create_address_scope(is_admin=True,
                                                   ip_version=4)
        # None-shared admin address scope cannot be retrieved by tenant user.
        self.assertRaises(lib_exc.NotFound, self.client.show_address_scope,
                          address_scope['id'])

    @decorators.attr(type='negative')
    @decorators.idempotent_id('5c25dc6a-1e92-467a-9cc7-cda74b6003db')
    def test_delete_non_existent_address_scope(self):
        non_exist_id = data_utils.rand_name('address_scope')
        self.assertRaises(lib_exc.NotFound, self.client.delete_address_scope,
                          non_exist_id)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('47c25dc5-e886-4a84-88c3-ac5031969661')
    def test_update_non_existent_address_scope(self):
        non_exist_id = data_utils.rand_name('address_scope')
        self.assertRaises(lib_exc.NotFound, self.client.update_address_scope,
                          non_exist_id, name='foo-name')

    @decorators.attr(type='negative')
    @decorators.idempotent_id('702d0515-82cb-4207-b0d9-703336e54665')
    def test_update_shared_address_scope_to_unshare(self):
        address_scope = self._create_address_scope(is_admin=True, shared=True,
                                                   ip_version=4)
        self.assertRaises(lib_exc.BadRequest,
                          self.admin_client.update_address_scope,
                          address_scope['id'], name='new-name', shared=False)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('1e471e5c-6f9c-437a-9257-fd9bc4b6f0fb')
    def test_delete_address_scope_associated_with_subnetpool(self):
        address_scope = self._create_address_scope(ip_version=4)
        prefixes = ['10.11.12.0/24']
        subnetpool_data = {
            'name': 'foo-subnetpool',
            'min_prefixlen': '29', 'prefixes': prefixes,
            'address_scope_id': address_scope['id']}
        self.create_subnetpool(**subnetpool_data)
        self.assertRaises(lib_exc.Conflict, self.client.delete_address_scope,
                          address_scope['id'])



# Copyright 2018 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr

from neutron_lib import constants as const
from tempest.common import utils
from tempest.lib.common.utils import data_utils
from tempest.lib import decorators
from tempest.lib import exceptions as tlib_exceptions

from neutron_tempest_plugin.api import base
from neutron_tempest_plugin import config

CONF = config.CONF


class PortTestCasesAdmin(base.BaseAdminNetworkTest):

    @classmethod
    def resource_setup(cls):
        super().resource_setup()
        cls.network = cls.create_network()
        cls.create_subnet(cls.network)

    @decorators.idempotent_id('dfe8cc79-18d9-4ae8-acef-3ec6bb719bb1')
    def test_update_mac_address(self):
        body = self.create_port(self.network)
        current_mac = body['mac_address']

        # Verify mac_address can be successfully updated.
        body = self.admin_client.update_port(body['id'],
                                             mac_address='12:34:56:78:be:6d')
        new_mac = body['port']['mac_address']
        self.assertNotEqual(current_mac, new_mac)
        self.assertEqual('12:34:56:78:be:6d', new_mac)

        # Verify that port update without specifying mac_address does not
        # change the mac address.
        body = self.admin_client.update_port(body['port']['id'],
                                             description='Port Description')
        self.assertEqual(new_mac, body['port']['mac_address'])

    @decorators.idempotent_id('dfe8cc79-18d9-4ae8-acef-3ec6bb719cc2')
    @utils.requires_ext(extension="port-mac-address-regenerate",
                        service="network")
    def test_regenerate_mac_address(self):
        body = self.create_port(self.network)
        current_mac = body['mac_address']
        body = self.admin_client.update_port(body['id'],
                                             mac_address=None)
        new_mac = body['port']['mac_address']
        self.assertNotEqual(current_mac, new_mac)
        self.assertTrue(netaddr.valid_mac(new_mac))

    @decorators.idempotent_id('4d75cc60-99d0-4949-b3ce-5826b81aa0a9')
    @utils.requires_ext(extension="port-trusted-vif",
                        service="network")
    def test_port_create_with_trusted_attr_set(self):
        port = self.admin_client.create_port(
            network_id=self.network['id'], trusted=True)['port']
        self.ports.append(port)
        self.assertTrue(port['trusted'])
        self.assertTrue(port['binding:profile']['trusted'])

        port = self.admin_client.create_port(
            network_id=self.network['id'], trusted=False)['port']
        self.ports.append(port)
        self.assertFalse(port['trusted'])
        self.assertFalse(port['binding:profile']['trusted'])

    @decorators.idempotent_id('26c15e2a-55b2-410f-8ed3-84db9406ff3f')
    @utils.requires_ext(extension="port-trusted-vif",
                        service="network")
    def test_port_set_trusted_attr(self):
        port = self.admin_client.create_port(
            network_id=self.network['id'])['port']
        self.ports.append(port)
        self.assertIsNone(port['trusted'])
        self.assertNotIn('trusted', port['binding:profile'])

        updated_port = self.admin_client.update_port(
            port['id'], trusted=True)['port']
        self.assertTrue(updated_port['trusted'])
        self.assertTrue(updated_port['binding:profile']['trusted'])

        updated_port = self.admin_client.update_port(
            port['id'], trusted=False)['port']
        self.assertFalse(updated_port['trusted'])
        self.assertFalse(updated_port['binding:profile']['trusted'])


class PortTestCasesResourceRequest(base.BaseAdminNetworkTest):

    required_extensions = ['port-resource-request',
                           'qos',
                           'qos-bw-minimum-ingress']

    EGRESS_KBPS = 1000
    INGRESS_KBPS = 2000
    ANY_KPPS = 500

    @classmethod
    def skip_checks(cls):
        super().skip_checks()
        if not config.CONF.neutron_plugin_options.provider_vlans:
            msg = "Skipped as provider VLANs are not available in config"
            raise cls.skipException(msg)

    @classmethod
    def resource_setup(cls):
        super().resource_setup()

        cls.vnic_type = 'normal'

        cls.network = cls.create_network()
        cls.physnet_name = CONF.neutron_plugin_options.provider_vlans[0]
        base_segm = CONF.neutron_plugin_options.provider_net_base_segm_id
        cls.prov_network = cls.create_provider_network(
            physnet_name=cls.physnet_name, start_segmentation_id=base_segm)

    @classmethod
    def setup_clients(cls):
        super().setup_clients()
        cls.qos_minimum_bandwidth_rules_client = \
            cls.os_admin.qos_minimum_bandwidth_rules_client
        cls.qos_bw_limit_rule_client = \
            cls.os_admin.qos_limit_bandwidth_rules_client
        cls.qos_minimum_packet_rate_rules_client = \
            cls.os_admin.qos_minimum_packet_rate_rules_client

    def _create_qos_policy_and_port(self, network, vnic_type,
                                    network_policy=False, min_kpps=False):
        qos_policy = self.create_qos_policy(
            name=data_utils.rand_name('test_policy'), shared=True)
        self.qos_minimum_bandwidth_rules_client.create_minimum_bandwidth_rule(
            qos_policy_id=qos_policy['id'],
            **{'direction': const.EGRESS_DIRECTION,
               'min_kbps': self.EGRESS_KBPS})

        self.qos_minimum_bandwidth_rules_client.create_minimum_bandwidth_rule(
            qos_policy_id=qos_policy['id'],
            **{'direction': const.INGRESS_DIRECTION,
               'min_kbps': self.INGRESS_KBPS})

        if min_kpps:
            self.qos_minimum_packet_rate_rules_client.\
                create_minimum_packet_rate_rule(
                    qos_policy_id=qos_policy['id'],
                    **{'direction': const.ANY_DIRECTION,
                    'min_kpps': min_kpps})

        port_policy_id = qos_policy['id'] if not network_policy else None
        port_kwargs = {
            'qos_policy_id': port_policy_id,
            'binding:vnic_type': vnic_type
        }

        if network_policy:
            self.admin_client.update_network(network['id'],
                                             qos_policy_id=qos_policy['id'])

        port_id = self.create_port(network, **port_kwargs)['id']
        return self.admin_client.show_port(port_id)['port']

    def _assert_resource_request(self, port, vnic_type, min_kpps=None):
        self.assertIn('resource_request', port)
        vnic_trait = 'CUSTOM_VNIC_TYPE_%s' % vnic_type.upper()
        physnet_trait = 'CUSTOM_PHYSNET_%s' % self.physnet_name.upper()
        if utils.is_extension_enabled('port-resource-request-groups',
                                      'network'):
            min_bw_group_found = False
            min_pps_group_found = False if min_kpps else True
            for rg in port['resource_request']['request_groups']:
                self.assertIn(rg['id'],
                              port['resource_request']['same_subtree'])
                if (('NET_BW_EGR_KILOBIT_PER_SEC' in rg['resources'] or
                        'NET_BW_IGR_KILOBIT_PER_SEC' in rg['resources']) and
                        not min_bw_group_found):
                    self.assertCountEqual([physnet_trait, vnic_trait],
                                          rg['required'])

                    self.assertEqual(
                        {'NET_BW_EGR_KILOBIT_PER_SEC': self.EGRESS_KBPS,
                        'NET_BW_IGR_KILOBIT_PER_SEC': self.INGRESS_KBPS},
                        rg['resources']
                    )
                    min_bw_group_found = True
                elif (('NET_PACKET_RATE_KILOPACKET_PER_SEC' in
                        rg['resources'] and min_kpps) and
                        not min_pps_group_found):
                    self.assertCountEqual([vnic_trait], rg['required'])

                    self.assertEqual(
                        {'NET_PACKET_RATE_KILOPACKET_PER_SEC': min_kpps},
                        rg['resources']
                    )
                    min_pps_group_found = True
                else:
                    self.fail('"resource_request" contains unexpected request '
                              'group: %s', rg)

            if not min_bw_group_found or not min_pps_group_found:
                self.fail('Did not find expected request groups in '
                          '"resource_request": %s',
                          port['resource_request']['request_groups'])
        else:
            self.assertCountEqual([physnet_trait, vnic_trait],
                                  port['resource_request']['required'])

            self.assertEqual(
                {'NET_BW_EGR_KILOBIT_PER_SEC': self.EGRESS_KBPS,
                'NET_BW_IGR_KILOBIT_PER_SEC': self.INGRESS_KBPS},
                port['resource_request']['resources']
            )

    @decorators.idempotent_id('ebb86dc4-716c-4558-8516-6dfc4a67601f')
    def test_port_resource_request(self):
        port = self._create_qos_policy_and_port(
            network=self.prov_network, vnic_type=self.vnic_type)
        port_id = port['id']

        self._assert_resource_request(port, self.vnic_type)

        # Note(lajoskatona): port-resource-request is an admin only feature,
        # so test if non-admin user can't see the new field.
        port = self.client.show_port(port_id)['port']
        self.assertNotIn('resource_request', port)

        self.update_port(port, **{'qos_policy_id': None})
        port = self.admin_client.show_port(port_id)['port']
        self.assertIsNone(port['resource_request'])

    @decorators.idempotent_id('5ae93aa0-408a-11ec-bbca-17b1a60f3438')
    @utils.requires_ext(service='network',
                        extension='port-resource-request-groups')
    def test_port_resource_request_min_bw_and_min_pps(self):
        port = self._create_qos_policy_and_port(
            network=self.prov_network, vnic_type=self.vnic_type,
            network_policy=False, min_kpps=self.ANY_KPPS)
        port_id = port['id']

        self._assert_resource_request(port, self.vnic_type,
                                      min_kpps=self.ANY_KPPS)

        # Note(lajoskatona): port-resource-request is an admin only feature,
        # so test if non-admin user can't see the new field.
        port = self.client.show_port(port_id)['port']
        self.assertNotIn('resource_request', port)

        self.update_port(port, **{'qos_policy_id': None})
        port = self.admin_client.show_port(port_id)['port']
        self.assertIsNone(port['resource_request'])

    @decorators.idempotent_id('7261391f-64cc-45a6-a1e3-435694c54bf5')
    def test_port_resource_request_no_provider_net_conflict(self):
        self.skipTest('This test is skipped until LP#1991965 is implemented. '
                      'Once implemented, it will be removed and new tests '
                      'added. For now it is temporarily kept as a reminder')
        conflict = self.assertRaises(
            tlib_exceptions.Conflict,
            self._create_qos_policy_and_port,
            network=self.network, vnic_type=self.vnic_type)
        self.assertEqual('QosRuleNotSupported', conflict.resp_body['type'])

    @decorators.idempotent_id('0eeb6ffa-9a7a-40b5-83dd-dbdcd67e2e64')
    def test_port_resource_request_empty(self):
        qos_policy = self.create_qos_policy(
            name=data_utils.rand_name('test_policy'), shared=True)

        # Note(lajoskatona): Add a non-minimum-bandwidth-rule to the policy
        # to make sure that the resource request is not filled with it.
        self.qos_bw_limit_rule_client.create_limit_bandwidth_rule(
            qos_policy['id'],
            **{'max_kbps': self.EGRESS_KBPS,
               'max_burst_kbps': 800,
               'direction': const.EGRESS_DIRECTION})

        port_kwargs = {
            'qos_policy_id': qos_policy['id'],
            'binding:vnic_type': self.vnic_type
        }

        port_id = self.create_port(self.prov_network, **port_kwargs)['id']
        port = self.admin_client.show_port(port_id)['port']

        self.assertIn('resource_request', port)
        self.assertIsNone(port['resource_request'])

    @decorators.idempotent_id('b6c34ae4-44c8-47f0-86de-7ef9866fa000')
    def test_port_resource_request_inherited_policy(self):
        base_segm = CONF.neutron_plugin_options.provider_net_base_segm_id
        prov_network = self.create_provider_network(
            physnet_name=self.physnet_name,
            start_segmentation_id=base_segm)
        port = self._create_qos_policy_and_port(
            network=prov_network, vnic_type=self.vnic_type,
            network_policy=True)

        self._assert_resource_request(port, self.vnic_type)



# Copyright 2023 Canonical
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import subprocess
import time
import typing

import netaddr
import testtools

from tempest.common import utils as tutils

from neutron_tempest_plugin import config
from neutron_tempest_plugin.scenario import base

from neutron_lib import constants as const

from oslo_log import log
from oslo_serialization import jsonutils

from os_ken.tests.integrated.common import docker_base as ctn_base

from tempest.lib.common.utils import data_utils
from tempest.lib import decorators
from tempest.lib import exceptions as lib_exc

CONF = config.CONF
LOG = log.getLogger(__name__)
FRR_BASE_IMAGE = 'quay.io/nf-core/ubuntu:22.04'


class FRROCIImage(ctn_base.DockerImage):
    def __init__(
        self,
        daemons: tuple[str, str],
        baseimage: str | None = None,
        use_existing: bool = False,
    ):
        super().__init__(baseimage=baseimage or FRR_BASE_IMAGE)
        self.daemons = daemons
        self.tagname = 'frr-' + '-'.join(daemons)
        if use_existing and self.exist(self.tagname):
            return

        workdir = os.path.join(ctn_base.TEST_BASE_DIR, self.tagname)
        pkgs = ' '.join(('telnet', 'tcpdump', 'frr'))
        c = ctn_base.CmdBuffer()
        c << f'FROM {self.baseimage}'
        c << 'RUN apt-get update'
        c << f'RUN apt-get install -qy --no-install-recommends {pkgs}'
        c << 'RUN echo "#!/bin/sh" > /frr'
        c << 'RUN echo mkdir -p /run/frr >> /frr'
        c << 'RUN echo chmod 755 /run/frr >> /frr'
        c << 'RUN echo chown frr:frr /run/frr >> /frr'
        c << (
            'RUN echo exec /usr/lib/frr/watchfrr '
            f'-F traditional {" ".join(self.daemons)}>> /frr'
        )
        c << 'RUN chmod +x /frr'
        c << 'CMD /frr'

        self.cmd.sudo(f'rm -rf {workdir}')
        self.cmd.execute(f'mkdir -p {workdir}')
        self.cmd.execute(f"echo '{str(c)}' > {workdir}/Dockerfile")
        self.build(self.tagname, workdir)


class FRRContainer(ctn_base.Container):
    class veth_info(typing.NamedTuple):
        bridge_name: str
        bridge_type: str
        ctn_ifname: str
        host_ifname: str

    _veths: list[veth_info]

    class route(typing.NamedTuple):
        dst: netaddr.IPNetwork
        next_hop: netaddr.IPNetwork

    _ctn_routes: list[route]

    def __init__(
        self,
        name: str,
        image: FRROCIImage,
    ):
        self._veths = []
        self._ctn_routes = []
        super().__init__(name, image.tagname)

    # XXX upstream to os-ken
    def next_if_name(self) -> str:
        name = f'eth{len(self.eths)}'
        self.eths.append(name)
        return name

    # XXX upstream to os-ken
    def run(self, network: str | None = None) -> int:
        c = ctn_base.CmdBuffer(' ')
        c << "docker run --privileged=true"
        for sv in self.shared_volumes:
            c << f"-v {sv[0]}:{sv[1]}"
        if network:
            c << f"--network {network}"
        c << "--name {0} --hostname {0} -id {1}".format(
            self.docker_name(), self.image
        )
        self.id = self.dcexec(str(c), retry=True)
        self.is_running = True
        self.exec_on_ctn("ip li set up dev lo")
        ipv4 = None
        ipv6 = None
        if network and network != 'none':
            ifname = self.next_if_name()
            for line in self.exec_on_ctn(f"ip a show dev {ifname}").split(
                '\n'
            ):
                if line.strip().startswith("inet "):
                    elems = [e.strip() for e in line.strip().split(' ')]
                    ipv4 = elems[1]
                elif line.strip().startswith("inet6 "):
                    elems = [e.strip() for e in line.strip().split(' ')]
                    ipv6 = elems[1]
            self.set_addr_info(
                bridge='docker0', ipv4=ipv4, ipv6=ipv6, ifname=ifname
            )
        return 0

    def wait_for_frr_daemons_up(
        self,
        try_times: int = 30,
        interval: int = 1,
    ) -> ctn_base.CommandOut:
        return self.cmd.sudo(
            f'docker logs {self.docker_name()} '
            '|grep "WATCHFRR.*all daemons up"',
            try_times=try_times,
            interval=interval,
        )

    @staticmethod
    def hash_ifname(ifname: str) -> str:
        # Assuming IFNAMSIZ of 16, with null-termination gives 15 characters.
        return 'veth' + str(hash(ifname) % 10**11)

    @staticmethod
    def get_if_mac(ifname: str) -> netaddr.EUI:
        with open(f'/sys/class/net/{ifname}/address') as faddr:
            return faddr.readline().rstrip()

    def add_veth_to_bridge(
        self,
        bridge_name: str,
        bridge_type: str,
        ipv4_cidr: str,
        ipv6_cidr: str,
        ipv6_prefix: netaddr.IPNetwork | None = None,
        vlan: int | None = None,
    ) -> None:
        if not self.is_running:
            LOG.error('The container must be running before calling '
                      'add_veth_to_bridge')
            raise
        if bridge_type != ctn_base.BRIDGE_TYPE_OVS:
            LOG.error('bridge_type must be %s', ctn_base.BRIDGE_TYPE_OVS)
            raise
        veth_pair = (
            self.hash_ifname(f'{self.name}-int{len(self._veths)}'),
            self.hash_ifname(f'{self.name}-ext{len(self._veths)}'),
        )
        self.cmd.sudo(
            f'ip link add {veth_pair[0]} type veth peer name {veth_pair[1]}'
        )
        if ipv6_prefix and not ipv6_cidr:
            eui = netaddr.EUI(self.get_if_mac(veth_pair[0]))
            ipv6_cidr = (
                f'{eui.ipv6(ipv6_prefix.first)}/{ipv6_prefix.prefixlen}'
            )

        self.cmd.sudo(f'ip link set netns {self.get_pid()} dev {veth_pair[0]}')
        self.cmd.sudo(f'ovs-vsctl add-port {bridge_name} {veth_pair[1]}')
        if vlan:
            self.cmd.sudo(f'ovs-vsctl set port {veth_pair[1]} tag={vlan}')

        ifname = self.next_if_name()
        self.exec_on_ctn(f'ip link set name {ifname} {veth_pair[0]}')

        # Ensure IPv6 is not disabled in container
        self.exec_on_ctn('sysctl -w net.ipv6.conf.all.disable_ipv6=0')

        for cidr in (ipv4_cidr, ipv6_cidr):
            if not cidr:
                continue
            self.exec_on_ctn(f'ip addr add {cidr} dev {ifname}')
        self.exec_on_ctn(f'ip link set up dev {ifname}')
        self.cmd.sudo(f'ip link set up dev {veth_pair[1]}')
        self.set_addr_info(
            bridge_name, ipv4=ipv4_cidr, ipv6=ipv6_cidr, ifname=ifname
        )
        self._veths.append(
            self.veth_info(
                bridge_name=bridge_name,
                bridge_type=bridge_type,
                ctn_ifname=ifname,
                host_ifname=veth_pair[1],
            )
        )

    def add_ctn_route(self, route: route) -> None:
        self.exec_on_ctn(
            f'ip -{route.dst.version} route add '
            f'{str(route.dst.cidr)} via {str(route.next_hop.ip)}'
        )
        self._ctn_routes.append(route)

    def del_ctn_route(self, route: route) -> None:
        self.exec_on_ctn(
            f'ip -{route.dst.version} route del '
            f'{str(route.dst.cidr)} via {str(route.next_hop.ip)}'
        )
        self._ctn_routes.remove(route)

    def remove(self, check_exist=True) -> ctn_base.CommandOut:
        for veth in self._veths:
            # The veth pair itself will be destroyed as a side effect of
            # removing the container, so we only need to clean up the bridge
            # attachment.
            if veth.bridge_type == ctn_base.BRIDGE_TYPE_BRCTL:
                self.cmd.sudo(
                    'brctl delif ' f'{veth.bridge_name} ' f'{veth.host_ifname}'
                )
            elif veth.bridge_type == ctn_base.BRIDGE_TYPE_OVS:
                self.cmd.sudo(
                    'ovs-vsctl del-port '
                    f'{veth.bridge_name} '
                    f'{veth.host_ifname}'
                )
        super().remove(check_exist=check_exist)

    def vtysh(self, cmd: list[str]) -> ctn_base.CommandOut:
        cmd_str = ' '.join(f"-c '{c}'" for c in cmd)
        return self.exec_on_ctn(f'vtysh {cmd_str}', capture=True)


class BFDContainer(FRRContainer):
    def __init__(
        self,
        name: str,
        image: FRROCIImage | None = None,
    ):
        image = image or FRROCIImage(
            daemons=('zebra', 'bfdd'), use_existing=True
        )
        super().__init__(name, image)
        if 'bfdd' not in image.daemons:
            LOG.error("'bffd' was not found in '%s'", image.daemons)
            raise

    def add_bfd_peer(self, ip_address: str) -> None:
        self.vtysh(
            [
                'enable',
                'conf',
                'bfd',
                f'peer {ip_address} interface eth0',
            ]
        )

    def del_bfd_peer(self, ip_address: str) -> None:
        self.vtysh(
            [
                'enable',
                'conf',
                'bfd',
                f'no peer {ip_address} interface eth0',
            ]
        )

    def show_bfd_peer(self, peer: str) -> dict[str, typing.Any]:
        return jsonutils.loads(self.vtysh([f'show bfd peer {peer} json']))

    def wait_for_bfd_peer_status(
        self, peer: str, status: str, try_times=30, interval=1
    ) -> None:
        while try_times:
            peer_data = self.show_bfd_peer(peer)
            if peer_data['status'] == status:
                return
            time.sleep(interval)
            try_times -= 1
        raise lib_exc.TimeoutException


class NetworkMultipleGWTest(base.BaseAdminTempestTestCase):
    """Test the following topology

    +------------------------------------------------------------------+
    |                          test runner                             |
    |                                                                  |
    |                                 +-----------+ eth0 public VLAN N |
    | +-------- br-ex ----------+     | FRR w/BFD |                    |
    | | +---------------------+ |     +-----------+ eth1 public flat   |
    | | |   public physnet    | |     +-----------+ eth0 public VLAN N |
    | | +---------------------+ |     | FRR w/BFD |                    |
    | +-------------------------+     +-----------+ eth1 public flat   |
    |     |              |                                             |
    +-----|--------------|---------------------------------------------+
          | -  VLAN N  - |
     +-------------------------+
     |      project router     | - enable_default_route_{bfd,ecmp}=True
     +-------------------------+
                 |
           +----------+
           | instance |
           +----------+

    NOTE(fnordahl) At the time of writing, FRR provides a BFD daemon, but has
    not integrated it with static routes [0][1].  As a consequence the
    test will manually add/remove routes on test runner to ensure correct path
    is chosen for traffic from test runner to instance.  On the return path the
    BFD implementation in OVN will ensure the correct path is chosen
    automatically.

    In real world usage most vendors have BFD support for static routes.

    0: https://github.com/FRRouting/frr/wiki/Feature-Requests
    1: https://github.com/FRRouting/frr/issues/3369
    """
    class host_route(typing.NamedTuple):
        dst: netaddr.IPNetwork
        next_hop: netaddr.IPNetwork

    host_routes: list[host_route] = []

    credentials = ['primary', 'admin']

    @classmethod
    def setup_clients(cls):
        super().setup_clients()
        if not cls.admin_client:
            cls.admin_client = cls.os_admin.network_client

    @classmethod
    @tutils.requires_ext(extension="external-gateway-multihoming",
                         service="network")
    def resource_setup(cls):
        super().resource_setup()

        # Ensure devstack configured public subnets are recorded, so that we
        # don't attempt to use them again.
        cls.reserve_external_subnet_cidrs()

        # We need to know prefixlength of the devstack configured public
        # subnets.
        for subnet_id in cls.admin_client.show_network(
                CONF.network.public_network_id)['network']['subnets']:
            subnet = cls.admin_client.show_subnet(subnet_id)['subnet']
            if subnet['ip_version'] == 4:
                cls.public_ipv4_subnet = subnet
                continue
            cls.public_ipv6_subnet = subnet
        cls.ext_networks = []
        for n in range(0, 2):
            ext_network = cls.create_provider_network(
                physnet_name='public',
                start_segmentation_id=4040 + n,
                external=True,
            )
            ext_ipv6_subnet = cls.create_subnet(
                ext_network,
                ip_version=const.IP_VERSION_6,
                client=cls.admin_client,
            )
            ext_ipv4_subnet = cls.create_subnet(
                ext_network,
                ip_version=const.IP_VERSION_4,
                client=cls.admin_client,
            )
            cls.ext_networks.append(
                (ext_network, ext_ipv6_subnet, ext_ipv4_subnet)
            )
        cls.host_routes = []
        cls.resource_setup_container()

    @classmethod
    def resource_setup_container(cls):
        cls.containers = []
        for n in range(0, 2):
            ext_network, ext_ipv6_subnet, ext_ipv4_subnet = cls.ext_networks[n]

            # frr container
            bfd_container = BFDContainer(data_utils.rand_name('frr'))
            cls.containers.append(bfd_container)

            bfd_container.run(network='none')
            public_ipv6_net = netaddr.IPNetwork(cls.public_ipv6_subnet['cidr'])
            public_ipv4_net = netaddr.IPNetwork(cls.public_ipv4_subnet['cidr'])
            ipv6_net = netaddr.IPNetwork(ext_ipv6_subnet['cidr'])
            ipv4_net = netaddr.IPNetwork(ext_ipv4_subnet['cidr'])
            # reserve an IP for container on the public network for routing
            # into the vlan network.
            fip_address = cls.create_floatingip()['floating_ip_address']
            cls.veths = [
                bfd_container.add_veth_to_bridge(
                    'br-ex',
                    ctn_base.BRIDGE_TYPE_OVS,
                    f'{ext_ipv4_subnet["gateway_ip"]}/{ipv4_net.prefixlen}',
                    f'{ext_ipv6_subnet["gateway_ip"]}/{ipv6_net.prefixlen}',
                    vlan=ext_network['provider:segmentation_id'],
                ),
                bfd_container.add_veth_to_bridge(
                    'br-ex',
                    ctn_base.BRIDGE_TYPE_OVS,
                    f'{fip_address}/{public_ipv4_net.prefixlen}',
                    '',
                    ipv6_prefix=public_ipv6_net,
                ),
            ]
            for subnet in (cls.public_ipv4_subnet, cls.public_ipv6_subnet):
                bfd_container.exec_on_ctn(
                    f'ip -{subnet["ip_version"]} route add default '
                    f'via {subnet["gateway_ip"]} dev eth1'
                )
            for ip_version in (6, 4):
                for addr_info in bfd_container.get_addr_info(
                    'br-ex', ip_version
                ).items():
                    if addr_info[1] == 'eth1':
                        if ip_version == 6:
                            dst_subnet = ext_ipv6_subnet
                        else:
                            dst_subnet = ext_ipv4_subnet
                        cls.add_host_route(
                            cls.host_routes,
                            cls.host_route(
                                netaddr.IPNetwork(dst_subnet["cidr"]),
                                netaddr.IPNetwork(addr_info[0]),
                            ),
                        )
            bfd_container.wait_for_frr_daemons_up()

    @classmethod
    def resource_cleanup(cls):
        # Ensure common cleanup code can clean up resources created by admin
        cls.client = cls.admin_client
        super().resource_cleanup()
        for ctn in cls.containers:
            try:
                ctn.stop()
            except ctn_base.CommandError:
                pass
            ctn.remove()
        # NOTE(fnordahl): the loop body modifies the list, so we need to
        # iterate on a copy.
        for route in cls.host_routes.copy():
            cls.del_host_route(cls.host_routes, route)

    @staticmethod
    def add_host_route(
        lst: list[host_route],
        route: host_route
    ) -> None:
        subprocess.run(
            (
                'sudo',
                'ip',
                f'-{route.dst.version}',
                'route',
                'add',
                str(route.dst.cidr),
                'via',
                str(route.next_hop.ip),
            ),
            capture_output=True,
            check=True,
            text=True,
        )
        lst.append(route)

    @staticmethod
    def del_host_route(
        lst: list[host_route],
        route: host_route
    ) -> None:
        subprocess.run(
            (
                'sudo',
                'ip',
                f'-{route.dst.version}',
                'route',
                'del',
                str(route.dst.cidr),
                'via',
                str(route.next_hop.ip),
            ),
            capture_output=True,
            check=True,
            text=True,
        )
        lst.remove(route)

    def add_ctn_route(
        self,
        ctn: BFDContainer,
        dst: netaddr.IPNetwork,
        next_hop: netaddr.IPNetwork,
    ):
        ctn_route = ctn.route(dst, next_hop)
        ctn.add_ctn_route(ctn_route)
        self.per_test_ctn_routes.append((ctn, ctn_route))

    def setUp(self):
        super().setUp()
        self.per_test_host_routes = []
        self.per_test_ctn_routes = []

    def tearDown(self):
        super().tearDown()
        # NOTE(fnordahl): the loop body modifies the list, so we need to
        # iterate on a copy.
        for ctn_route in self.per_test_ctn_routes.copy():
            ctn = ctn_route[0]
            route = ctn_route[1]
            ctn.del_ctn_route(route)
        for host_route in self.per_test_host_routes.copy():
            self.del_host_route(self.per_test_host_routes, host_route)

    def add_routes_for_router(
        self,
        router: dict[str, typing.Any],
        ctn: FRRContainer,
        add_ctn_route: bool = True,
        add_host_route: bool = True,
    ):
        ports = self._list_router_interfaces(
            self.admin_client,
            router_id=router['id']
        )
        for port in ports:
            if port['device_owner'] != const.DEVICE_OWNER_ROUTER_INTF:
                continue
            for fixed_ip in port['fixed_ips']:
                subnet = self.client.show_subnet(
                    fixed_ip['subnet_id'])['subnet']
                for addr_info in ctn.get_addr_info(
                    'br-ex',
                    subnet['ip_version'],
                ).items():
                    if addr_info[1] == 'eth0':
                        # container route
                        ctn_net = netaddr.IPNetwork(addr_info[0])
                        for gw_info in router['external_gateways']:
                            for ip_info in gw_info['external_fixed_ips']:
                                if (
                                    ip_info['ip_address'] in ctn_net and
                                    add_ctn_route
                                ):
                                    self.add_ctn_route(
                                        ctn,
                                        netaddr.IPNetwork(subnet['cidr']),
                                        netaddr.IPNetwork(
                                            ip_info['ip_address']
                                        ),
                                    )
                    elif addr_info[1] == 'eth1' and add_host_route:
                        self.add_host_route(
                            self.per_test_host_routes,
                            self.host_route(
                                netaddr.IPNetwork(self.subnet['cidr']),
                                netaddr.IPNetwork(addr_info[0]),
                            ),
                        )

    @testtools.skipUnless(
        CONF.compute.min_compute_nodes == 1,
        'More than 1 compute node, test only works on '
        'single node configurations.',
    )
    @decorators.idempotent_id('9baa05e6-ba10-4850-93e3-695f4d97b8f8')
    def test_create_router_single_gw_bfd(self):
        ext_network_id = self.ext_networks[0][0]['id']
        bfd_container = self.containers[0]
        router = self.create_admin_router(
            router_name=data_utils.rand_name('router'),
            admin_state_up=True,
            enable_snat=False,
            enable_default_route_bfd=True,
            external_network_id=ext_network_id,
        )
        self.assertTrue(router['enable_default_route_bfd'])

        # Add BFD peers on bfd_container.
        for gw_info in router['external_gateways']:
            for ip_info in gw_info['external_fixed_ips']:
                bfd_container.add_bfd_peer(ip_info["ip_address"])
                bfd_container.wait_for_bfd_peer_status(
                    ip_info['ip_address'], 'up'
                )

        self.setup_network_and_server(
            router=router,
            create_fip=False,
            router_client=self.admin_client,
        )

        self.add_routes_for_router(router, bfd_container)

        # check connectivity
        self.check_connectivity(
            self.port['fixed_ips'][0]['ip_address'],
            CONF.validation.image_ssh_user,
            self.keypair['private_key'],
        )

    @testtools.skipUnless(
        CONF.compute.min_compute_nodes == 1,
        'More than 1 compute node, test only works on '
        'single node configurations.',
    )
    @decorators.idempotent_id('75202251-c384-4962-8685-60cf2c530906')
    def test_update_router_single_gw_bfd(self):
        ext_network_id = self.ext_networks[0][0]['id']
        bfd_container = self.containers[0]
        router = self._create_router_with_client(
            self.admin_client,
            router_name=data_utils.rand_name('router'),
            admin_state_up=True,
            enable_snat=False,
            external_network_id=ext_network_id,
        )
        self.assertFalse(router['enable_default_route_bfd'])

        self.setup_network_and_server(
            router=router,
            create_fip=False,
            router_client=self.admin_client,
        )

        self.add_routes_for_router(router, bfd_container)

        # check connectivity
        self.check_connectivity(
            self.port['fixed_ips'][0]['ip_address'],
            CONF.validation.image_ssh_user,
            self.keypair['private_key'],
        )

        # Enable BFD on router.
        #
        # NOTE(fnordahl): We need to repeat the `enable_snat` state, otherwise
        # the state will be toggled to the default value of 'True'.
        router = self.admin_client.update_router_with_snat_gw_info(
            router['id'],
            enable_snat=False,
            enable_default_route_bfd=True,
        )['router']
        self.assertTrue(router['enable_default_route_bfd'])

        # Add BFD peers on bfd_container.
        for gw_info in router['external_gateways']:
            for ip_info in gw_info['external_fixed_ips']:
                bfd_container.add_bfd_peer(ip_info["ip_address"])
                bfd_container.wait_for_bfd_peer_status(
                    ip_info['ip_address'], 'up'
                )

        # check connectivity
        self.check_connectivity(
            self.port['fixed_ips'][0]['ip_address'],
            CONF.validation.image_ssh_user,
            self.keypair['private_key'],
        )

    @testtools.skipUnless(
        CONF.compute.min_compute_nodes == 1,
        'More than 1 compute node, test only works on '
        'single node configurations.',
    )
    @decorators.idempotent_id('5117587d-9633-48b7-aa8f-ec9d59a601a5')
    def test_create_router_multiple_gw_bfd_and_ecmp(self):
        router = self.create_admin_router(
            router_name=data_utils.rand_name('router'),
            admin_state_up=True,
            enable_default_route_bfd=True,
            enable_default_route_ecmp=True,
        )
        router = self.admin_client.router_add_external_gateways(
            router['id'],
            [
                {
                    'network_id': self.ext_networks[0][0]['id'],
                    'enable_snat': False,
                },
                {
                    'network_id': self.ext_networks[1][0]['id'],
                    'enable_snat': False,
                },
            ],
        )['router']

        self.setup_network_and_server(
            router=router,
            create_fip=False,
            router_client=self.admin_client,
        )

        # Add BFD peers on bfd_containers.
        for gw_info in router['external_gateways']:
            for ip_info in gw_info['external_fixed_ips']:
                ip = netaddr.IPAddress(ip_info['ip_address'])
                for ctn in self.containers:
                    for addr_info in ctn.get_addr_info(
                        'br-ex',
                        ip.version,
                    ).items():
                        if addr_info[1] == 'eth0':
                            ctn_net = netaddr.IPNetwork(addr_info[0])
                            if ip not in ctn_net:
                                break
                            ctn.add_bfd_peer(str(ip))
                            ctn.wait_for_bfd_peer_status(str(ip), 'up')

        # Add route to project network on all containers.
        for ctn in self.containers:
            self.add_routes_for_router(router, ctn, True, False)

        # Add host route to project network via FRR container and confirm
        # connectivity one by one.
        #
        # We deliberately don't add both host routes at once as that would be
        # testing test runner configuration and linux kernel ECMP, which is out
        # of scope for our test.
        for ctn in self.containers:
            self.add_routes_for_router(router, ctn, False, True)

            # check connectivity
            self.check_connectivity(
                self.port['fixed_ips'][0]['ip_address'],
                CONF.validation.image_ssh_user,
                self.keypair['private_key'],
            )
            for host_route in self.per_test_host_routes.copy():
                self.del_host_route(self.per_test_host_routes, host_route)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr
from tempest.lib import decorators

from neutron_tempest_plugin.api import base


class SubnetsSearchCriteriaTest(base.BaseSearchCriteriaTest):

    resource = 'subnet'

    list_kwargs = {'shared': False}

    @classmethod
    def resource_setup(cls):
        if 'subnet-external-network' in cls.get_loaded_network_extensions():
            cls.list_kwargs['router:external'] = False
        super().resource_setup()
        net = cls.create_network(network_name='subnet-search-test-net')
        for name in cls.resource_names:
            cls.create_subnet(net, name=name)

    @decorators.idempotent_id('d2d61995-5dd5-4b93-bce7-3edefdb79563')
    def test_list_sorts_asc(self):
        self._test_list_sorts_asc()

    @decorators.idempotent_id('c3c6b0af-c4ac-4da0-b568-8d08ae550604')
    def test_list_sorts_desc(self):
        self._test_list_sorts_desc()

    @decorators.idempotent_id('b93063b3-f713-406e-bf93-e5738e09153c')
    def test_list_pagination(self):
        self._test_list_pagination()

    @decorators.idempotent_id('2ddd9aa6-de28-410f-9cbc-ce752893c407')
    def test_list_pagination_with_marker(self):
        self._test_list_pagination_with_marker()

    @decorators.idempotent_id('351183ef-6ed9-4d71-a9f2-a5ac049bd7ea')
    def test_list_pagination_with_href_links(self):
        self._test_list_pagination_with_href_links()

    @decorators.idempotent_id('dfaa20ca-6d84-4f26-962f-2fee4d247cd9')
    def test_list_pagination_page_reverse_asc(self):
        self._test_list_pagination_page_reverse_asc()

    @decorators.idempotent_id('40552213-3e12-4d6a-86f3-dda92f3de88c')
    def test_list_pagination_page_reverse_desc(self):
        self._test_list_pagination_page_reverse_desc()

    @decorators.idempotent_id('3cea9053-a731-4480-93ee-19b2c28a9ce4')
    def test_list_pagination_page_reverse_with_href_links(self):
        self._test_list_pagination_page_reverse_with_href_links()

    @decorators.idempotent_id('d851937c-9821-4b46-9d18-43e9077ecac0')
    def test_list_no_pagination_limit_0(self):
        self._test_list_no_pagination_limit_0()

    @decorators.idempotent_id('c0f9280b-9d81-4728-a967-6be22659d4c8')
    def test_list_validation_filters(self):
        self._test_list_validation_filters(self.list_kwargs)
        self._test_list_validation_filters({
            'unknown_filter': 'value'}, filter_is_valid=False)


class SubnetServiceTypeTestJSON(base.BaseNetworkTest):

    required_extensions = ['subnet-service-types']

    @classmethod
    def resource_setup(cls):
        super().resource_setup()
        cls.network = cls.create_network()

    @decorators.idempotent_id('7e0edb66-1bb2-4473-ab83-d039cddced0d')
    def test_allocate_ips_are_from_correct_subnet(self):
        cidr_1 = netaddr.IPNetwork('192.168.1.0/24')
        cidr_2 = netaddr.IPNetwork('192.168.2.0/24')

        # NOTE(slaweq): service_type "network:distributed" is needed for
        # ML2/OVN backend. It's needed because OVN driver creates additional
        # port for metadata service in each subnet with enabled dhcp and such
        # port needs to have allocated IP address from the subnet also.
        self.create_subnet(
            self.network,
            service_types=['test:type_1', 'network:distributed'],
            cidr=str(cidr_1))
        self.create_subnet(
            self.network,
            service_types=['test:type_2', 'network:distributed'],
            cidr=str(cidr_2))
        port_type_1 = self.create_port(self.network,
                                       device_owner="test:type_1")
        port_type_2 = self.create_port(self.network,
                                       device_owner="test:type_2")

        self.assertEqual(1, len(port_type_1['fixed_ips']))
        self.assertEqual(1, len(port_type_2['fixed_ips']))
        self.assertIn(
            netaddr.IPAddress(port_type_1['fixed_ips'][0]['ip_address']),
            cidr_1)
        self.assertIn(
            netaddr.IPAddress(port_type_2['fixed_ips'][0]['ip_address']),
            cidr_2)



# Copyright 2022 Troila
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import constants
from tempest.lib.common.utils import data_utils
from tempest.lib import decorators
from tempest.lib import exceptions

from neutron_tempest_plugin.api import base
from neutron_tempest_plugin import config

CONF = config.CONF


class NDPProxyTestJSON(base.BaseAdminNetworkTest):

    credentials = ['primary', 'admin']
    required_extensions = ['router', 'l3-ndp-proxy', 'address-scope']

    @classmethod
    def resource_setup(cls):
        super().resource_setup()
        address_scope = cls.create_address_scope(
            "test-as", **{'ip_version': constants.IP_VERSION_6})
        subnetpool = cls.create_subnetpool(
            "test-subnetpool",
            **{'address_scope_id': address_scope['id'],
               'default_prefixlen': 112,
               'prefixes': ['2001:abc::0/96']})
        # Create an external network and it's subnet
        ext_net = cls.create_network('test-ext-net', client=cls.admin_client,
                                     external=True)
        cls.create_subnet(
            ext_net, client=cls.admin_client,
            ip_version=constants.IP_VERSION_6,
            **{'subnetpool_id': subnetpool['id'], "cidr": "2001:abc::1:0/112"})
        # Create network, subnet, router and add interface
        cls.network = cls.create_network()
        cls.subnet = cls.create_subnet(
            cls.network, ip_version=constants.IP_VERSION_6,
            **{'subnetpool_id': subnetpool['id'], "cidr": "2001:abc::2:0/112"})
        cls.router = cls.create_router(data_utils.rand_name('router'),
                                       external_network_id=ext_net['id'],
                                       enable_ndp_proxy=True)
        cls.create_router_interface(cls.router['id'], cls.subnet['id'])

    @decorators.idempotent_id('481bc712-d504-4128-bffb-62d98b88886b')
    def test_ndp_proxy_lifecycle(self):
        port = self.create_port(self.network)
        np_description = 'Test ndp proxy description'
        np_name = 'test-ndp-proxy'

        #  Create ndp proxy
        created_ndp_proxy = self.create_ndp_proxy(
            name=np_name,
            description=np_description,
            router_id=self.router['id'],
            port_id=port['id'])
        self.assertEqual(self.router['id'], created_ndp_proxy['router_id'])
        self.assertEqual(port['id'], created_ndp_proxy['port_id'])
        self.assertEqual(np_description, created_ndp_proxy['description'])
        self.assertEqual(np_name, created_ndp_proxy['name'])
        self.assertEqual(port['fixed_ips'][0]['ip_address'],
                         created_ndp_proxy['ip_address'])

        # Show created ndp_proxy
        body = self.client.get_ndp_proxy(created_ndp_proxy['id'])
        ndp_proxy = body['ndp_proxy']
        self.assertEqual(np_description, ndp_proxy['description'])
        self.assertEqual(self.router['id'], ndp_proxy['router_id'])
        self.assertEqual(port['id'], ndp_proxy['port_id'])
        self.assertEqual(np_name, ndp_proxy['name'])
        self.assertEqual(port['fixed_ips'][0]['ip_address'],
                         ndp_proxy['ip_address'])

        # List ndp proxies
        body = self.client.list_ndp_proxies()
        ndp_proxy_ids = [np['id'] for np in body['ndp_proxies']]
        self.assertIn(created_ndp_proxy['id'], ndp_proxy_ids)

        # Update ndp proxy
        updated_ndp_proxy = self.client.update_ndp_proxy(
                               created_ndp_proxy['id'],
                               name='updated_ndp_proxy')
        self.assertEqual('updated_ndp_proxy',
                         updated_ndp_proxy['ndp_proxy']['name'])
        self.assertEqual(
            np_description, updated_ndp_proxy['ndp_proxy']['description'])
        self.assertEqual(self.router['id'],
                         updated_ndp_proxy['ndp_proxy']['router_id'])
        self.assertEqual(port['id'], updated_ndp_proxy['ndp_proxy']['port_id'])
        self.assertEqual(port['fixed_ips'][0]['ip_address'],
                         updated_ndp_proxy['ndp_proxy']['ip_address'])

        # Delete ndp proxy
        self.delete_ndp_proxy(created_ndp_proxy)
        self.assertRaises(exceptions.NotFound,
                          self.client.get_ndp_proxy, created_ndp_proxy['id'])



# Copyright 2022 Troila
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import constants
from tempest.lib.common.utils import data_utils
from tempest.lib import decorators
from tempest.lib import exceptions

from neutron_tempest_plugin.api import base
from neutron_tempest_plugin import config

CONF = config.CONF


class NDPProxyNegativeTestJSON(base.BaseAdminNetworkTest):

    credentials = ['primary', 'admin']
    required_extensions = ['router', 'l3-ndp-proxy', 'address-scope']

    @classmethod
    def resource_setup(cls):
        super().resource_setup()
        address_scope = cls.create_address_scope(
            "test-as", **{'ip_version': constants.IP_VERSION_6})
        subnetpool = cls.create_subnetpool(
            "test-subnetpool",
            **{'address_scope_id': address_scope['id'],
               'default_prefixlen': 112,
               'prefixes': ['2001:abc::0/96']})
        # Create an external network and it's subnet
        ext_net = cls.create_network('test-ext-net', client=cls.admin_client,
                                     external=True)
        cls.create_subnet(
            ext_net, client=cls.admin_client,
            ip_version=constants.IP_VERSION_6,
            **{'subnetpool_id': subnetpool['id'], "cidr": "2001:abc::1:0/112"})
        # Create network, subnet, router and add interface
        cls.network = cls.create_network()
        cls.subnet = cls.create_subnet(
            cls.network, ip_version=constants.IP_VERSION_6,
            cidr='2002::abcd:0/112')
        cls.router = cls.create_router(
            data_utils.rand_name('router'),
            external_network_id=ext_net['id'])

    @decorators.attr(type='negative')
    @decorators.idempotent_id('a0897204-bb85-41cc-a5fd-5d0ab8116a07')
    def test_enable_ndp_proxy_without_external_gw(self):
        self.client.update_router(self.router['id'], external_gateway_info={})
        self.assertRaises(exceptions.Conflict,
            self.client.update_router,
            self.router['id'],
            enable_ndp_proxy=True)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('26e534a0-3e47-4894-8cb5-20a078ce76a9')
    def test_create_ndp_proxy_with_subnet_not_connect_router(self):
        self.client.update_router(self.router['id'], enable_ndp_proxy=True)
        port = self.create_port(self.network)
        self.assertRaises(exceptions.Conflict,
            self.create_ndp_proxy,
            self.router['id'],
            port_id=port['id'])

    @decorators.attr(type='negative')
    @decorators.idempotent_id('a0d93fd6-1219-4b05-9db8-bdf02846c447')
    def test_create_ndp_proxy_with_different_address_scope(self):
        self.client.update_router(self.router['id'], enable_ndp_proxy=True)
        address_scope = self.create_address_scope(
            "test-as", ip_version=constants.IP_VERSION_6)
        subnet_pool = self.create_subnetpool(
            "test-sp", address_scope_id=address_scope['id'],
            prefixes=['2002::abc:0/112'], default_prefixlen=112)
        network = self.create_network()
        subnet = self.create_subnet(
            network, ip_version=constants.IP_VERSION_6,
            cidr="2002::abc:0/112", subnetpool_id=subnet_pool['id'],
            reserve_cidr=False)
        self.create_router_interface(self.router['id'], subnet['id'])
        port = self.create_port(network)
        self.assertRaises(exceptions.Conflict, self.create_ndp_proxy,
                          self.router['id'], port_id=port['id'])

    @decorators.attr(type='negative')
    @decorators.idempotent_id('f9a4e56d-3836-40cd-8c05-585b3f1e034a')
    def test_create_ndp_proxy_without_ipv6_address(self):
        self.client.update_router(
            self.router['id'], enable_ndp_proxy=True)
        subnet = self.create_subnet(
            self.network, ip_version=constants.IP_VERSION_4)
        self.create_router_interface(self.router['id'], subnet['id'])
        port = self.create_port(self.network)
        self.assertRaises(exceptions.Conflict,
                          self.create_ndp_proxy,
                          self.router['id'], port_id=port['id'])

    @decorators.attr(type='negative')
    @decorators.idempotent_id('e035b3af-ebf9-466d-9ef5-a73b063a1f56')
    def test_enable_ndp_proxy_and_unset_gateway(self):
        self.assertRaises(exceptions.Conflict,
                          self.client.update_router,
                          self.router['id'],
                          enable_ndp_proxy=True,
                          external_gateway_info={})

    @decorators.attr(type='negative')
    @decorators.idempotent_id('194b5ee7-4c59-4643-aabf-80a125c3f688')
    def test_enable_ndp_proxy_without_address_scope(self):
        extnet = self.create_network("extnet", client=self.admin_client,
                                     external=True)
        self.create_subnet(extnet, client=self.admin_client,
                           ip_version=constants.IP_VERSION_6,
                           cidr='2001:abc1::0/112')
        self.assertRaises(exceptions.Conflict,
                          self.client.create_router,
                          name=data_utils.rand_name('router'),
                          enable_ndp_proxy=True,
                          external_gateway_info={'network_id': extnet['id']})
        router = self.create_router(data_utils.rand_name('router'))
        self.assertRaises(exceptions.Conflict,
                          self.client.update_router,
                          router['id'], enable_ndp_proxy=True,
                          external_gateway_info={'network_id': extnet['id']})



# Copyright (c) 2020 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from tempest.lib.common.utils import data_utils
from tempest.lib import decorators

from neutron_tempest_plugin.common import ip
from neutron_tempest_plugin.common import ssh
from neutron_tempest_plugin import config
from neutron_tempest_plugin.scenario import base


LOG = logging.getLogger(__name__)
CONF = config.CONF
MIN_VLAN_ID = 1
MAX_VLAN_ID = 4094


class BaseVlanTest(base.BaseAdminTempestTestCase):

    """Base class common for the tests for the "vlan_transparent" and "qinq".

    This base class covers common things for the tests for networks with
    enabled either `qinq` or `vlan_transparent` attributes. Those 2 attributes
    are functionally the same even from the end user's point of view. The only
    difference between them is ethtype used for the outer VLAN tag but this
    can't really be tested in the neutron-tempest-plugin tests.
    """

    network_kwargs = {}

    @classmethod
    def resource_setup(cls):
        super().resource_setup()
        # setup basic topology for servers we can log into
        cls.rand_name = data_utils.rand_name(
            cls.__name__.rsplit('.', 1)[-1])
        cls.network = cls.create_network(name=cls.rand_name,
                                         **cls.network_kwargs)
        cls.subnet = cls.create_subnet(network=cls.network,
                                       name=cls.rand_name)
        cls.router = cls.create_router_by_client()
        cls.create_router_interface(cls.router['id'], cls.subnet['id'])
        cls.keypair = cls.create_keypair(name=cls.rand_name)
        cls.vm_ports = []
        cls.security_group = cls.create_security_group(name=cls.rand_name)
        cls.create_loginable_secgroup_rule(cls.security_group['id'])

        cls.setup_advanced_image()

    @classmethod
    def skip_checks(cls):
        super().skip_checks()
        if not (CONF.neutron_plugin_options.advanced_image_ref or
                CONF.neutron_plugin_options.default_image_is_advanced):
            raise cls.skipException(
                'Advanced image is required to run these tests.')

    def _create_port_and_server(self, index,
                                port_security=True,
                                allowed_address_pairs=None):
        server_name = 'server-%s-%d' % (self.rand_name, index)
        port_name = 'port-%s-%d' % (self.rand_name, index)
        if port_security:
            sec_groups = [self.security_group['id']]
        else:
            sec_groups = None
        self.vm_ports.append(
            self.create_port(network=self.network, name=port_name,
                             security_groups=sec_groups,
                             port_security_enabled=port_security,
                             allowed_address_pairs=allowed_address_pairs))
        return self.create_server(flavor_ref=self.flavor_ref,
                                  image_ref=self.image_ref,
                                  key_name=self.keypair['name'],
                                  networks=[{'port': self.vm_ports[-1]['id']}],
                                  name=server_name)['server']

    def _configure_inner_vlan(self, port, ssh_client, vlan_tag, vlan_ip):
        ip_command = ip.IPCommand(ssh_client=ssh_client)
        addresses = ip_command.list_addresses(port=port)
        port_iface = ip.get_port_device_name(addresses, port)
        subport_iface = ip_command.configure_inner_vlan(
            port=port, vlan_tag=vlan_tag, ip_addresses=[vlan_ip])

        for address in ip_command.list_addresses(ip_addresses=vlan_ip):
            self.assertEqual(subport_iface, address.device.name)
            self.assertEqual(port_iface, address.device.parent)
            break
        else:
            self.fail("Sub-port fixed IP not found on server.")

    def _create_ssh_client(self, floating_ip):
        return ssh.Client(host=floating_ip['floating_ip_address'],
                          username=self.username,
                          pkey=self.keypair['private_key'])

    def _test_basic_inner_vlan_connectivity(
            self, port_security=True, use_allowed_address_pairs=False):
        self._ensure_ethtype()
        vlan_tag = data_utils.rand_int_id(start=MIN_VLAN_ID, end=MAX_VLAN_ID)
        vtag = vlan_tag % 256
        vlan_ipmask_template = '192.%d.%d.{ip_last_byte}/24' % (vtag, vtag)
        vms = []
        vlan_ipmasks = []
        floating_ips = []
        ssh_clients = []

        for i in range(2):
            vlan_ipmasks.append(vlan_ipmask_template.format(
                ip_last_byte=(i + 1) * 10))
            if use_allowed_address_pairs:
                allowed_address_pairs = [{'ip_address': vlan_ipmasks[i]}]
            else:
                allowed_address_pairs = None
            vms.append(self._create_port_and_server(
                index=i,
                port_security=port_security,
                allowed_address_pairs=allowed_address_pairs))
            floating_ips.append(self.create_floatingip(port=self.vm_ports[-1]))
            ssh_clients.append(
                self._create_ssh_client(floating_ip=floating_ips[i]))

            self.check_connectivity(ssh_client=ssh_clients[i])
            self._configure_inner_vlan(port=self.vm_ports[-1],
                                       ssh_client=ssh_clients[i],
                                       vlan_tag=vlan_tag,
                                       vlan_ip=vlan_ipmasks[i])

        if port_security:
            # Ping from vm0 to vm1 via VLAN interface should fail because
            # we haven't allowed ICMP
            self.check_remote_connectivity(
                ssh_clients[0],
                vlan_ipmasks[1].split('/')[0],
                servers=vms,
                should_succeed=False)

            # allow intra-security-group traffic
            sg_rule = self.create_pingable_secgroup_rule(
                self.security_group['id'])
            self.addCleanup(
                    self.os_primary.network_client.delete_security_group_rule,
                    sg_rule['id'])

        # Ping from vm0 to vm1 via VLAN interface should pass because
        # either port security is disabled or the ICMP sec group rule has been
        # added
        self.check_remote_connectivity(
            ssh_clients[0],
            vlan_ipmasks[1].split('/')[0],
            servers=vms)
        # Ping from vm1 to vm0 and check untagged packets are not dropped
        self.check_remote_connectivity(
            ssh_clients[1],
            self.vm_ports[-2]['fixed_ips'][0]['ip_address'],
            servers=vms)


class VlanTransparencyTest(BaseVlanTest):
    credentials = ['primary', 'admin']
    force_tenant_isolation = False

    required_extensions = ['vlan-transparent', 'allowed-address-pairs']

    network_kwargs = {'vlan_transparent': True}

    def _ensure_ethtype(self):
        self.assertTrue(self.network.get('vlan_transparent'))
        self.assertFalse(self.network.get('qinq'))

    @decorators.idempotent_id('a2694e3a-6d4d-4a23-9fcc-c3ed3ef37b16')
    def test_vlan_transparent_port_sec_disabled(self):
        self._test_basic_inner_vlan_connectivity(
            port_security=False, use_allowed_address_pairs=False)

    @decorators.idempotent_id('2dd03b4f-9c20-4cda-8c6a-40fa453ec69a')
    def test_vlan_transparent_allowed_address_pairs(self):
        self._test_basic_inner_vlan_connectivity(
            port_security=True, use_allowed_address_pairs=True)


class VlanQinqTest(BaseVlanTest):
    credentials = ['primary', 'admin']
    force_tenant_isolation = False

    required_extensions = ['qinq', 'allowed-address-pairs']

    network_kwargs = {
        'qinq': True,
        'provider_network_type': 'vlan'}

    @classmethod
    def resource_setup(cls):
        cls.network_kwargs['provider_physical_network'] = (
            CONF.neutron_plugin_options.provider_vlans[0])
        super().resource_setup()

    @classmethod
    def skip_checks(cls):
        super().skip_checks()
        if not CONF.neutron_plugin_options.provider_vlans:
            raise cls.skipException(
                'Physical network is required to run these tests.')

    def _ensure_ethtype(self):
        self.assertFalse(self.network.get('vlan_transparent'))
        self.assertTrue(self.network.get('qinq'))

    @decorators.idempotent_id('ae78398e-9242-46b4-a5fc-227581821fca')
    def test_vlan_transparent_port_sec_disabled(self):
        self._test_basic_inner_vlan_connectivity(
            port_security=False, use_allowed_address_pairs=False)

    @decorators.idempotent_id('6ca983cd-b1c5-4e2c-949e-4be8ffa22a9c')
    def test_vlan_transparent_allowed_address_pairs(self):
        self._test_basic_inner_vlan_connectivity(
            port_security=True, use_allowed_address_pairs=True)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils

from tempest.common import utils
from tempest.lib import decorators
from tempest.lib import exceptions as lib_exc

from neutron_tempest_plugin.tap_as_a_service import base


class TapMirrorTestJSON(base.BaseTaasTest):

    @classmethod
    @utils.requires_ext(extension='tap-mirror', service='network')
    def skip_checks(cls):
        super().skip_checks()

    @classmethod
    def resource_setup(cls):
        super().resource_setup()
        cls.network = cls.create_network()
        cls.tap_mirror_port = cls.create_port(cls.network)
        cls.in_direction = {'IN': 101}
        cls.out_direction = {'OUT': 102}
        cls.both_direction = cls.in_direction | cls.out_direction
        cls.remote_ip = '192.101.0.42'
        cls.remote_ip2 = '192.101.3.43'
        cls.gre = 'gre'
        cls.erspan = 'erspanv1'

    @decorators.idempotent_id('628f202c-ed0a-4eb1-8547-4954f67a84b7')
    def test_create_tap_mirror(self):
        tap_mirror = self.create_tap_mirror(
            port_id=self.tap_mirror_port['id'],
            directions=self.in_direction,
            remote_ip=self.remote_ip,
            mirror_type=self.gre
        )
        self.assertEqual(self.tap_mirror_port['id'], tap_mirror['port_id'])
        self.assertEqual('gre', tap_mirror['mirror_type'])
        self.assertEqual(self.in_direction, tap_mirror['directions'])
        self.tap_mirrors_client.delete_tap_mirror(tap_mirror['id'])

        tap_mirror = self.create_tap_mirror(
            port_id=self.tap_mirror_port['id'],
            directions=self.both_direction,
            remote_ip=self.remote_ip,
            mirror_type=self.erspan
        )
        self.assertEqual(self.tap_mirror_port['id'], tap_mirror['port_id'])
        self.assertEqual(self.erspan, tap_mirror['mirror_type'])
        self.assertEqual(self.both_direction, tap_mirror['directions'])

    @decorators.idempotent_id('299c251b-e0bc-4449-98db-959a5d8038c2')
    def test_list_show_tap_mirror(self):
        tap_mirror = self.create_tap_mirror(
            port_id=self.tap_mirror_port['id'],
            directions=self.out_direction,
            remote_ip=self.remote_ip,
            mirror_type=self.gre
        )
        tap_mirrors = self.tap_mirrors_client.list_tap_mirrors()
        is_t_m_found = False
        for t_m in tap_mirrors['tap_mirrors']:
            if t_m['id'] == tap_mirror['id']:
                is_t_m_found = True
                break
        self.assertTrue(is_t_m_found)
        tap_mirror_show_res = self.tap_mirrors_client.show_tap_mirror(
            tap_mirror['id'])['tap_mirror']
        self.assertEqual(tap_mirror['id'], tap_mirror_show_res['id'])
        self.assertEqual(self.gre, tap_mirror_show_res['mirror_type'])
        self.assertEqual(self.remote_ip,
                         tap_mirror_show_res['remote_ip'])
        self.assertEqual(self.out_direction,
                         tap_mirror_show_res['directions'])

    @decorators.idempotent_id('19c40379-bda5-48c9-8873-fc990739d1b5')
    def test_update_tap_mirror(self):
        tap_mirror = self.create_tap_mirror(
            port_id=self.tap_mirror_port['id'],
            directions=self.in_direction,
            remote_ip=self.remote_ip,
            mirror_type=self.gre
        )
        self.tap_mirrors_client.update_tap_mirror(
            tap_mirror_id=tap_mirror['id'],
            name='new_name',
            description='My fancy Tap Mirror'
        )
        tap_mirror_show_res = self.tap_mirrors_client.show_tap_mirror(
            tap_mirror['id'])['tap_mirror']
        self.assertEqual('new_name', tap_mirror_show_res['name'])
        self.assertEqual('My fancy Tap Mirror',
                         tap_mirror_show_res['description'])

    @decorators.idempotent_id('9ed165af-7c54-43ac-b14f-077e8f9601f6')
    def test_delete_mirror_port_deletes_tap_mirror(self):
        port1 = self.create_port(self.network)
        tap_mirror = self.create_tap_mirror(
            port_id=port1['id'],
            directions=self.out_direction,
            remote_ip=self.remote_ip,
            mirror_type=self.gre
        )
        # Delete port will result in deteltion of the tap_mirror
        self.ports_client.delete_port(port1['id'])
        self.assertRaises(lib_exc.NotFound,
                          self.tap_mirrors_client.show_tap_mirror,
                          tap_mirror['id'])

    @decorators.idempotent_id('abdd4451-bd9d-4f1e-ab7f-e949b9246714')
    def test_delete_tap_mirror_port_remains(self):
        port1 = self.create_port(self.network)
        tap_mirror = self.create_tap_mirror(
            port_id=port1['id'],
            directions=self.out_direction,
            remote_ip=self.remote_ip,
            mirror_type=self.gre
        )
        # Delete tap_mirror will keep the port
        self.tap_mirrors_client.delete_tap_mirror(tap_mirror['id'])
        port_res = self.ports_client.show_port(port1['id'])['port']
        self.assertEqual(port1['name'], port_res['name'])

    @decorators.idempotent_id('1d8b68fc-a600-4b9e-bd17-9469c3a6c95b')
    def test_create_tap_mirror_negative(self):
        # directions keys' valid values are IN and OUT
        self.assertRaises(lib_exc.BadRequest,
                          self.create_tap_mirror,
                          port_id=self.tap_mirror_port['id'],
                          directions={'something': 101},
                          remote_ip=self.remote_ip,
                          mirror_type=self.gre)
        # mirror_type valid values are erspanv1 and gre
        self.assertRaises(lib_exc.BadRequest,
                          self.create_tap_mirror,
                          port_id=self.tap_mirror_port['id'],
                          directions=self.out_direction,
                          remote_ip=self.remote_ip,
                          mirror_type='erspanv2')
        # remote_ip must be a valid IP
        self.assertRaises(lib_exc.BadRequest,
                          self.create_tap_mirror,
                          port_id=self.tap_mirror_port['id'],
                          directions=self.in_direction,
                          remote_ip='192.101.0.420',
                          mirror_type=self.gre)

    @decorators.idempotent_id('2b7850b3-3920-4f16-96b7-05e2efd96877')
    def test_create_tap_service_tunnel_id_conflict(self):
        self.create_tap_mirror(
            port_id=self.tap_mirror_port['id'],
            directions=self.in_direction,
            remote_ip=self.remote_ip,
            mirror_type=self.gre
        )

        port2 = self.create_port(self.network)
        self.addCleanup(self.ports_client.delete_port, port2['id'])
        self.assertRaises(lib_exc.Conflict,
                          self.create_tap_mirror,
                          port_id=port2['id'],
                          directions=self.in_direction,
                          remote_ip='192.101.0.4',
                          mirror_type=self.gre)

    @decorators.idempotent_id('95ef1cc1-cd57-4193-a88e-716795e39ebf')
    def test_create_tap_mirror_non_existing_port(self):
        not_exists = uuidutils.generate_uuid()
        self.assertRaises(lib_exc.NotFound,
                          self.create_tap_mirror,
                          port_id=not_exists,
                          directions=self.out_direction,
                          remote_ip=self.remote_ip,
                          mirror_type=self.gre)

    @decorators.idempotent_id('123202cd-d810-4c15-bae7-26d69b24a1a4')
    def test_multiple_mirrors_for_port(self):
        port1 = self.create_port(self.network)
        tap_mirror = self.create_tap_mirror(
            port_id=port1['id'],
            directions=self.out_direction,
            remote_ip=self.remote_ip,
            mirror_type=self.gre
        )
        self.addCleanup(self.tap_mirrors_client.delete_tap_mirror,
                        tap_mirror['id'])

        # Creation of the 2nd mirror in case the direction: tunnel_id dict
        # is different.
        tap_mirror2 = self.create_tap_mirror(
            port_id=port1['id'],
            directions={'OUT': 103},
            remote_ip=self.remote_ip2,
            mirror_type=self.gre
        )

        # We have a conflict if the direction: tunnel_id dict is the
        # same
        self.tap_mirrors_client.delete_tap_mirror(tap_mirror2['id'])
        self.assertRaises(lib_exc.Conflict,
                          self.create_tap_mirror,
                          port_id=port1['id'],
                          directions=self.out_direction,
                          remote_ip='192.101.0.4',
                          mirror_type=self.gre)



# Copyright 2018 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr
from neutron_lib import constants
from oslo_log import log
from tempest.lib.common.utils import data_utils
from tempest.lib import decorators

from neutron_tempest_plugin.common import ip
from neutron_tempest_plugin.common import ssh
from neutron_tempest_plugin.common import utils
from neutron_tempest_plugin import config
from neutron_tempest_plugin import exceptions
from neutron_tempest_plugin.scenario import base


CONF = config.CONF
LOG = log.getLogger(__name__)
PYTHON3_BIN = "python3"


def get_receiver_script(group, port, hello_message, ack_message, result_file):

    return """
import socket
import struct
import sys

multicast_group = '%(group)s'
server_address = ('', %(port)s)

# Create the socket
sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)

# Bind to the server address
sock.bind(server_address)

# Tell the operating system to add the socket to the multicast group
# on all interfaces.
group = socket.inet_aton(multicast_group)
mreq = struct.pack('4sL', group, socket.INADDR_ANY)
sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)

# Receive/respond loop
with open('%(result_file)s', 'w') as f:
    f.write('%(hello_message)s')
    f.flush()
    data, address = sock.recvfrom(1024)
    f.write('received ' + str(len(data)) + ' bytes from ' + str(address))
    f.write(str(data))
sock.sendto(b'%(ack_message)s', address)
    """ % {'group': group,
           'port': port,
           'hello_message': hello_message,
           'ack_message': ack_message,
           'result_file': result_file}


def get_sender_script(group, port, message, result_file):

    return """
import socket
import sys

message = b'%(message)s'
multicast_group = ('%(group)s', %(port)s)

# Create the datagram socket
sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
# Set the time-to-live for messages to 1 so they do not go past the
# local network segment.
sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 1)

# Set a timeout so the socket does not block indefinitely when trying
# to receive data.
sock.settimeout(1)

with open('%(result_file)s', 'w') as f:
    try:
        # Send data to the multicast group
        sent = sock.sendto(message, multicast_group)

        # Look for responses from all recipients
        while True:
            try:
                data, server = sock.recvfrom(1024)
            except socket.timeout:
                f.write('timed out, no more responses')
                break
            else:
                f.write('received reply ' + str(data) + ' from ' + str(server))
    finally:
        sys.stdout.write('closing socket')
        sock.close()
    """ % {'group': group,
           'port': port,
           'message': message,
           'result_file': result_file}


def get_unregistered_script(interface, group, result_file):
    return """#!/bin/bash
export LC_ALL=en_US.UTF-8
tcpdump -i %(interface)s host %(group)s -vvneA -s0 -l -c1 &> %(result_file)s &
    """ % {'interface': interface,
           'group': group,
           'result_file': result_file}


class BaseMulticastTest:

    credentials = ['primary', 'admin']
    force_tenant_isolation = False

    # Import configuration options
    available_type_drivers = (
        CONF.neutron_plugin_options.available_type_drivers)

    hello_message = "I am waiting..."
    multicast_port = 5007
    multicast_message = "Big Bang"
    receiver_output_file = "/tmp/receiver_mcast_out"
    sender_output_file = "/tmp/sender_mcast_out"
    unregistered_output_file = "/tmp/unregistered_mcast_out"

    @classmethod
    def skip_checks(cls):
        super().skip_checks()
        advanced_image_available = (
            CONF.neutron_plugin_options.advanced_image_ref or
            CONF.neutron_plugin_options.default_image_is_advanced)
        if not advanced_image_available:
            skip_reason = "This test require advanced tools for this test"
            raise cls.skipException(skip_reason)

    @classmethod
    def resource_setup(cls):
        super().resource_setup()

        cls.setup_advanced_image()

        # setup basic topology for servers we can log into it
        cls.network = cls.create_network()
        cls.subnet = cls.create_subnet(cls.network)
        cls.router = cls.create_router_by_client()
        cls.create_router_interface(cls.router['id'], cls.subnet['id'])

        cls.keypair = cls.create_keypair()

        cls.secgroup = cls.os_primary.network_client.create_security_group(
            name='secgroup_mtu')
        cls.security_groups.append(cls.secgroup['security_group'])
        cls.create_loginable_secgroup_rule(
            secgroup_id=cls.secgroup['security_group']['id'])
        cls.create_pingable_secgroup_rule(
            secgroup_id=cls.secgroup['security_group']['id'])
        # Create security group rule for UDP (multicast traffic)
        cls.create_secgroup_rules(
            rule_list=[dict(protocol=constants.PROTO_NAME_UDP,
                            direction=constants.INGRESS_DIRECTION,
                            remote_ip_prefix=cls.any_addresses,
                            ethertype=cls.ethertype)],
            secgroup_id=cls.secgroup['security_group']['id'])

        # Multicast IP range to be used for multicast group IP asignement
        if '-' in cls.multicast_group_range:
            multicast_group_range = netaddr.IPRange(
                *cls.multicast_group_range.split('-'))
        else:
            multicast_group_range = netaddr.IPNetwork(
                cls.multicast_group_range)
        cls.multicast_group_iter = iter(multicast_group_range)

    def _create_server(self):
        name = data_utils.rand_name("multicast-server")
        server = self.create_server(
            flavor_ref=self.flavor_ref,
            image_ref=self.image_ref,
            key_name=self.keypair['name'], name=name,
            networks=[{'uuid': self.network['id']}],
            security_groups=[{'name': self.secgroup['security_group']['name']}]
        )['server']
        self.wait_for_server_active(server)
        self.wait_for_guest_os_ready(server)
        server['port'] = self.client.list_ports(
            network_id=self.network['id'], device_id=server['id'])['ports'][0]
        server['fip'] = self.create_floatingip(port=server['port'])
        server['ssh_client'] = ssh.Client(server['fip']['floating_ip_address'],
                                          self.username,
                                          pkey=self.keypair['private_key'])
        self._check_cmd_installed_on_server(server['ssh_client'],
                                            server, PYTHON3_BIN)
        return server

    def _prepare_sender(self, server, mcast_address):
        check_script = get_sender_script(
            group=mcast_address, port=self.multicast_port,
            message=self.multicast_message,
            result_file=self.sender_output_file)
        server['ssh_client'].execute_script(
            'echo "%s" > /tmp/multicast_traffic_sender.py' % check_script)

    def _prepare_receiver(self, server, mcast_address):
        check_script = get_receiver_script(
            group=mcast_address, port=self.multicast_port,
            hello_message=self.hello_message, ack_message=server['id'],
            result_file=self.receiver_output_file)
        ssh_client = ssh.Client(
            server['fip']['floating_ip_address'],
            self.username,
            pkey=self.keypair['private_key'])
        self._check_cmd_installed_on_server(ssh_client, server,
                                            PYTHON3_BIN)
        server['ssh_client'].execute_script(
            'echo "%s" > /tmp/multicast_traffic_receiver.py' % check_script)

    def _prepare_unregistered(self, server, mcast_address):
        ssh_client = ssh.Client(
            server['fip']['floating_ip_address'],
            self.username,
            pkey=self.keypair['private_key'])
        ip_command = ip.IPCommand(ssh_client=ssh_client)
        addresses = ip_command.list_addresses(port=server['port'])
        port_iface = ip.get_port_device_name(addresses, server['port'])
        check_script = get_unregistered_script(
            interface=port_iface, group=mcast_address,
            result_file=self.unregistered_output_file)
        self._check_cmd_installed_on_server(ssh_client, server,
                                            'tcpdump')
        server['ssh_client'].execute_script(
            'echo "%s" > /tmp/unregistered_traffic_receiver.sh' % check_script)

    @decorators.idempotent_id('113486fc-24c9-4be4-8361-03b1c9892867')
    def test_multicast_between_vms_on_same_network(self):
        """Test multicast messaging between two servers on the same network

        [Sender server] -> (Multicast network) -> [Receiver server]
        """
        LOG.debug("IGMP snooping enabled: %s",
                  CONF.neutron_plugin_options.is_igmp_snooping_enabled)
        sender = self._create_server()
        receivers = [self._create_server() for _ in range(1)]
        # Sender can be also receiver of multicast traffic
        receivers.append(sender)
        unregistered = self._create_server()
        self._check_multicast_conectivity(sender=sender, receivers=receivers,
                                          unregistered=unregistered)

    def _is_multicast_traffic_expected(self, mcast_address):
        """Checks if multicast traffic is expected to arrive.

        Checks if multicast traffic is expected to arrive to the
        unregistered VM.

        If IGMP snooping is enabled, multicast traffic should not be
        flooded unless the destination IP is in the range of 224.0.0.X
        [0].

        [0] https://tools.ietf.org/html/rfc4541 (See section 2.1.2)
        """
        return (str(mcast_address).startswith('224.0.0') or not
                CONF.neutron_plugin_options.is_igmp_snooping_enabled)

    def _check_multicast_conectivity(self, sender, receivers, unregistered):
        """Test multi-cast messaging between two servers

        [Sender server] -> ... some network topology ... -> [Receiver server]
        """
        mcast_address = next(self.multicast_group_iter)
        LOG.debug("Multicast group address: %s", mcast_address)

        def _message_received(client, msg, file_path):
            result = client.execute_script(
                "cat {path} || echo '{path} not exists yet'".format(
                    path=file_path))
            return msg in result

        self._prepare_unregistered(unregistered, mcast_address)

        # Run the unregistered node script
        unregistered['ssh_client'].execute_script(
            "bash /tmp/unregistered_traffic_receiver.sh", become_root=True)

        self._prepare_sender(sender, mcast_address)
        receiver_ids = []
        for receiver in receivers:
            self._prepare_receiver(receiver, mcast_address)
            receiver['ssh_client'].execute_script(
                "%s /tmp/multicast_traffic_receiver.py &" % PYTHON3_BIN,
                shell="bash")
            utils.wait_until_true(
                lambda: _message_received(
                    receiver['ssh_client'], self.hello_message,
                    self.receiver_output_file),
                exception=RuntimeError(
                    "Receiver script didn't start properly on server "
                    "{!r}.".format(receiver['id'])))

            receiver_ids.append(receiver['id'])

        # Now lets run scripts on sender
        sender['ssh_client'].execute_script(
            "%s /tmp/multicast_traffic_sender.py" % PYTHON3_BIN)

        # And check if message was received
        for receiver in receivers:
            utils.wait_until_true(
                lambda: _message_received(
                    receiver['ssh_client'], self.multicast_message,
                    self.receiver_output_file),
                exception=RuntimeError(
                    "Receiver {!r} didn't get multicast message".format(
                        receiver['id'])))

        def _sender_completed():
            replies_result = sender['ssh_client'].execute_script(
                "cat {path} 2>/dev/null || echo ''".format(
                    path=self.sender_output_file))
            for receiver_id in receiver_ids:
                expected_pattern = "received reply b'{}' from".format(
                    receiver_id)
                if expected_pattern not in replies_result:
                    return False
            return replies_result.count('received reply') == len(receiver_ids)

        utils.wait_until_true(
            _sender_completed,
            exception=RuntimeError("Sender didn't complete properly"))

        def check_unregistered_host():
            unregistered_result = unregistered['ssh_client'].execute_script(
                "cat {path} || echo '{path} not exists yet'".format(
                    path=self.unregistered_output_file))
            LOG.debug("Unregistered VM result: %s", unregistered_result)
            return expected_result in unregistered_result

        expected_result = '1 packet captured'
        unregistered_error_message = (
            'Unregistered server did not received expected packet.')
        if not self._is_multicast_traffic_expected(mcast_address):
            # Kill the tcpdump command runs on the unregistered node with "-c"
            # option so it will be stopped automatically if it will receive
            # packet matching filters,
            # We don't expect any packets to be captured really in this case
            # so let's kill tcpdump so it flushes its output to the output
            # file.
            expected_result = ('0 packets captured')
            unregistered_error_message = (
                'Unregistered server received unexpected packet(s).')
            try:
                unregistered['ssh_client'].execute_script(
                    "killall tcpdump && sleep 2", become_root=True)
            except exceptions.SSHScriptFailed:
                # Probably some packet was captured by tcpdump and due to that
                # it is already stopped
                self.assertTrue(check_unregistered_host(),
                                unregistered_error_message)
                return

        utils.wait_until_true(
            check_unregistered_host,
            exception=RuntimeError(unregistered_error_message))


class MulticastTestIPv4(BaseMulticastTest, base.BaseTempestTestCase):

    # Import configuration options
    multicast_group_range = CONF.neutron_plugin_options.multicast_group_range

    # IP version specific parameters
    _ip_version = constants.IP_VERSION_4
    any_addresses = constants.IPv4_ANY



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.lib import decorators
from tempest.lib import exceptions as lib_exc
import testtools

from neutron_tempest_plugin.api import base
from tempest import config

CONF = config.CONF


class NetworksNegativeTest(base.BaseNetworkTest):

    @classmethod
    def resource_setup(cls):
        super().resource_setup()
        cls.network = cls.create_network()
        cls.subnet = cls.create_subnet(cls.network)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('9f80f25b-5d1b-4f26-9f6b-774b9b270819')
    def test_delete_network_in_use(self):
        self.create_port(self.network)
        with testtools.ExpectedException(lib_exc.Conflict):
            self.client.delete_subnet(self.subnet['id'])
        with testtools.ExpectedException(lib_exc.Conflict):
            self.client.delete_network(self.network['id'])

    @decorators.attr(type='negative')
    @decorators.idempotent_id('9f80f25b-5d1b-4f26-9f6b-774b9b270820')
    def test_update_network_mtu(self):
        with testtools.ExpectedException(lib_exc.BadRequest):
            self.client.create_network(
                mtu=CONF.neutron_plugin_options.max_mtu + 1)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('53537bba-d6c3-4a2e-bda4-ab5b009fb7d9')
    def test_create_subnet_mtu_below_minimum_ipv4(self):
        network = self.create_network(mtu=67)
        with testtools.ExpectedException(lib_exc.Conflict):
            self.create_subnet(network, ip_version=4, cidr='10.0.0.0/24')

    @decorators.attr(type='negative')
    @decorators.idempotent_id('1de68cb6-e6d4-47df-b820-c5048796f33a')
    @testtools.skipUnless(config.CONF.network_feature_enabled.ipv6,
                          'IPv6 is not enabled')
    def test_create_subnet_mtu_below_minimum_ipv6(self):
        network = self.create_network(mtu=1279)
        with testtools.ExpectedException(lib_exc.Conflict):
            self.create_subnet(network, ip_version=6, cidr='2001:db8:0:1::/64')

    @decorators.attr(type='negative')
    @decorators.idempotent_id('5213df6d-7141-40b2-90ea-a958d9bc97e5')
    def test_update_network_mtu_below_minimum_ipv4(self):
        network = self.create_network(mtu=1280)
        self.create_subnet(network, ip_version=4, cidr='10.0.0.0/24')
        with testtools.ExpectedException(lib_exc.Conflict):
            self.client.update_network(network['id'], mtu=67)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('1a714fc4-24b1-4c07-a005-d5c218672eab')
    @testtools.skipUnless(config.CONF.network_feature_enabled.ipv6,
                          'IPv6 is not enabled')
    def test_update_network_mtu_below_minimum_ipv6(self):
        network = self.create_network(mtu=1280)
        self.create_subnet(network, ip_version=6, cidr='2001:db8:0:1::/64')
        with testtools.ExpectedException(lib_exc.Conflict):
            self.client.update_network(network['id'], mtu=1279)



# Copyright (c) 2017 Midokura SARL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time

import ddt
from neutron_lib import constants as lib_constants
from neutron_lib.services.qos import constants as qos_consts
from oslo_log import log
from tempest.common import utils
from tempest.common import waiters
from tempest.lib.common.utils import data_utils
from tempest.lib import decorators
from tempest.lib import exceptions
import testtools

from neutron_tempest_plugin.api import base as base_api
from neutron_tempest_plugin.common import ssh
from neutron_tempest_plugin.common import utils as common_utils
from neutron_tempest_plugin import config
from neutron_tempest_plugin.scenario import base
from neutron_tempest_plugin.scenario import constants
from neutron_tempest_plugin.scenario import test_qos


CONF = config.CONF
LOG = log.getLogger(__name__)


class FloatingIpTestCasesMixin:
    credentials = ['primary', 'admin']

    @classmethod
    @utils.requires_ext(extension="router", service="network")
    def resource_setup(cls):
        super().resource_setup()
        cls.network = cls.create_network()
        cls.subnet = cls.create_subnet(cls.network)
        cls.router = cls.create_router_by_client()
        cls.create_router_interface(cls.router['id'], cls.subnet['id'])
        cls.keypair = cls.create_keypair()

        cls.secgroup = cls.os_primary.network_client.create_security_group(
            name=data_utils.rand_name('secgroup'))['security_group']
        cls.security_groups.append(cls.secgroup)
        cls.create_loginable_secgroup_rule(secgroup_id=cls.secgroup['id'])
        cls.create_pingable_secgroup_rule(secgroup_id=cls.secgroup['id'])

        if cls.same_network:
            cls._dest_network = cls.network
        else:
            cls._dest_network = cls._create_dest_network()

    @classmethod
    def _get_external_gateway(cls):
        if CONF.network.public_network_id:
            subnets = cls.os_admin.network_client.list_subnets(
                network_id=CONF.network.public_network_id)

            for subnet in subnets['subnets']:
                if (subnet['gateway_ip'] and
                    subnet['ip_version'] == lib_constants.IP_VERSION_4):
                    return subnet['gateway_ip']

    @classmethod
    def _create_dest_network(cls):
        network = cls.create_network()
        subnet = cls.create_subnet(network)
        cls.create_router_interface(cls.router['id'], subnet['id'])
        return network

    def _create_server(self, create_floating_ip=True, network=None):
        if network is None:
            network = self.network
        port = self.create_port(network, security_groups=[self.secgroup['id']])
        if create_floating_ip:
            fip = self.create_floatingip(port=port)
        else:
            fip = None
        server = self.create_server(
            flavor_ref=CONF.compute.flavor_ref,
            image_ref=CONF.compute.image_ref,
            key_name=self.keypair['name'],
            networks=[{'port': port['id']}])['server']
        waiters.wait_for_server_status(self.os_primary.servers_client,
                                       server['id'],
                                       constants.SERVER_STATUS_ACTIVE)
        return {'port': port, 'fip': fip, 'server': server}

    def _test_east_west(self, src_has_fip, dest_has_fip):
        # The proxy VM is used to control the source VM when it doesn't
        # have a floating-ip.
        if src_has_fip:
            proxy = None
            proxy_client = None
        else:
            proxy = self._create_server()
            proxy_client = ssh.Client(proxy['fip']['floating_ip_address'],
                                      CONF.validation.image_ssh_user,
                                      pkey=self.keypair['private_key'])

        # Source VM
        if src_has_fip:
            src_server = self._create_server()
            src_server_ip = src_server['fip']['floating_ip_address']
        else:
            src_server = self._create_server(create_floating_ip=False)
            src_server_ip = src_server['port']['fixed_ips'][0]['ip_address']
        ssh_client = ssh.Client(src_server_ip,
                                CONF.validation.image_ssh_user,
                                pkey=self.keypair['private_key'],
                                proxy_client=proxy_client)

        # Destination VM
        if dest_has_fip:
            dest_server = self._create_server(network=self._dest_network)
        else:
            dest_server = self._create_server(create_floating_ip=False,
                                              network=self._dest_network)

        # Check connectivity
        self.check_remote_connectivity(ssh_client,
            dest_server['port']['fixed_ips'][0]['ip_address'],
            servers=[src_server, dest_server])
        if dest_has_fip:
            self.check_remote_connectivity(ssh_client,
                dest_server['fip']['floating_ip_address'],
                servers=[src_server, dest_server])


@ddt.ddt
class FloatingIpSameNetwork(FloatingIpTestCasesMixin,
                            base.BaseTempestTestCase):

    same_network = True

    @decorators.idempotent_id('05c4e3b3-7319-4052-90ad-e8916436c23b')
    @ddt.unpack
    @ddt.data({'src_has_fip': True, 'dest_has_fip': True},
              {'src_has_fip': True, 'dest_has_fip': False},
              {'src_has_fip': False, 'dest_has_fip': True},
              {'src_has_fip': True, 'dest_has_fip': False})
    def test_east_west(self, src_has_fip, dest_has_fip):
        self._test_east_west(src_has_fip=src_has_fip,
                             dest_has_fip=dest_has_fip)


@ddt.ddt
class FloatingIpSeparateNetwork(FloatingIpTestCasesMixin,
                                base.BaseTempestTestCase):

    same_network = False

    @decorators.idempotent_id('f18f0090-3289-4783-b956-a0f8ac511e8b')
    @ddt.unpack
    @ddt.data({'src_has_fip': True, 'dest_has_fip': True},
              {'src_has_fip': True, 'dest_has_fip': False},
              {'src_has_fip': False, 'dest_has_fip': True},
              {'src_has_fip': True, 'dest_has_fip': False})
    def test_east_west(self, src_has_fip, dest_has_fip):
        self._test_east_west(src_has_fip=src_has_fip,
                             dest_has_fip=dest_has_fip)


class DefaultSnatToExternal(FloatingIpTestCasesMixin,
                            base.BaseTempestTestCase):
    same_network = True

    @decorators.idempotent_id('3d73ea1a-27c6-45a9-b0f8-04a283d9d764')
    def test_snat_external_ip(self):
        """Check connectivity to an external IP"""
        gateway_external_ip = self._get_external_gateway()

        if not gateway_external_ip:
            raise self.skipTest("IPv4 gateway is not configured for public "
                                "network or public_network_id is not "
                                "configured")
        proxy = self._create_server()
        proxy_client = ssh.Client(proxy['fip']['floating_ip_address'],
                                  CONF.validation.image_ssh_user,
                                  pkey=self.keypair['private_key'])
        src_server = self._create_server(create_floating_ip=False)
        src_server_ip = src_server['port']['fixed_ips'][0]['ip_address']
        ssh_client = ssh.Client(src_server_ip,
                                CONF.validation.image_ssh_user,
                                pkey=self.keypair['private_key'],
                                proxy_client=proxy_client)
        self.check_remote_connectivity(ssh_client,
                                       gateway_external_ip,
                                       servers=[proxy, src_server])

    def _test_nested_snat_external_ip(self, feature_enabled_bool):
        """Check connectivity to an external IP from a nested network."""
        gateway_external_ip = self._get_external_gateway()

        if not gateway_external_ip:
            raise self.skipTest("IPv4 gateway is not configured for public "
                                "network or public_network_id is not "
                                "configured")
        proxy = self._create_server()
        proxy_client = ssh.Client(proxy['fip']['floating_ip_address'],
                                  CONF.validation.image_ssh_user,
                                  pkey=self.keypair['private_key'])

        # Create a nested router
        router = self.create_router(
            router_name=data_utils.rand_name('router'),
            admin_state_up=True)

        # Attach outer subnet to it
        outer_port = self.create_port(self.network)
        self.client.add_router_interface_with_port_id(router['id'],
                                                      outer_port['id'])

        # Attach a nested subnet to it
        network = self.create_network()
        subnet = self.create_subnet(network)
        self.create_router_interface(router['id'], subnet['id'])

        # Set up static routes in both directions
        self.client.update_extra_routes(
            self.router['id'],
            outer_port['fixed_ips'][0]['ip_address'], subnet['cidr'])
        self.client.update_extra_routes(
            router['id'], self.subnet['gateway_ip'], '0.0.0.0/0')

        # Create a server inside the nested network
        src_server = self._create_server(create_floating_ip=False,
                                         network=network)

        # Check connectivity if nested SNAT is enabled, else no connectivity
        src_server_ip = src_server['port']['fixed_ips'][0]['ip_address']
        ssh_client = ssh.Client(src_server_ip,
                                CONF.validation.image_ssh_user,
                                pkey=self.keypair['private_key'],
                                proxy_client=proxy_client)
        self.check_remote_connectivity(ssh_client,
                                       gateway_external_ip,
                                       should_succeed=feature_enabled_bool,
                                       servers=[proxy, src_server])

    @decorators.idempotent_id('b911b124-b6cb-449d-83d9-b34f3665741d')
    @utils.requires_ext(extension='extraroute', service='network')
    def test_nested_snat_external_ip(self):
        feature_enabled_bool = (
            CONF.neutron_plugin_options.snat_rules_apply_to_nested_networks
        )
        self._test_nested_snat_external_ip(feature_enabled_bool)


class FloatingIPPortDetailsTest(FloatingIpTestCasesMixin,
                                base.BaseTempestTestCase):
    same_network = True

    @classmethod
    @utils.requires_ext(extension="router", service="network")
    @utils.requires_ext(extension="fip-port-details", service="network")
    def resource_setup(cls):
        super().resource_setup()

    @decorators.idempotent_id('a663aeee-dd81-492b-a207-354fd6284dbe')
    def test_floatingip_port_details(self):
        """Tests the following:

        1. Create a port with floating ip in Neutron.
        2. Create two servers in Nova.
        3. Attach the port to the server.
        4. Detach the port from the server.
        5. Attach the port to the second server.
        6. Detach the port from the second server.
        """
        port = self.create_port(self.network)
        fip = self.create_and_associate_floatingip(port['id'])
        server1 = self._create_server(create_floating_ip=False)
        server2 = self._create_server(create_floating_ip=False)

        for server in [server1, server2]:
            # attach the port to the server
            self.create_interface(
                server['server']['id'], port_id=port['id'])
            waiters.wait_for_interface_status(
                self.os_primary.interfaces_client, server['server']['id'],
                port['id'], lib_constants.PORT_STATUS_ACTIVE)
            fip = self.client.show_floatingip(fip['id'])['floatingip']
            server_data = self.os_admin.servers_client.show_server(
                server['server']['id'])['server']
            zone = 'compute:' + server_data['OS-EXT-AZ:availability_zone']
            self._check_port_details(
                fip, port, status=lib_constants.PORT_STATUS_ACTIVE,
                device_id=server['server']['id'],
                device_owner=zone)
            LOG.debug('Port check for server %s and FIP %s finished, '
                      'lets detach port %s from server!',
                      server['server']['id'], fip['id'], port['id'])

            # detach the port from the server; this is a cast in the compute
            # API so we have to poll the port until the device_id is unset.
            self.delete_interface(server['server']['id'], port['id'])
            port = self._wait_for_port_detach(port['id'])
            LOG.debug('Port %s has been detached from server %s, lets check '
                      'the status of port in FIP %s details!',
                      port['id'], server['server']['id'], fip['id'])
            fip = self._wait_for_fip_port_down(fip['id'])
            self._check_port_details(
                fip, port, status=lib_constants.PORT_STATUS_DOWN,
                device_id='', device_owner='')

    def _check_port_details(self, fip, port, status, device_id, device_owner):
        self.assertIn('port_details', fip)
        port_details = fip['port_details']
        self.assertEqual(port['name'], port_details['name'])
        self.assertEqual(port['network_id'], port_details['network_id'])
        self.assertEqual(port['mac_address'], port_details['mac_address'])
        self.assertEqual(port['admin_state_up'],
                         port_details['admin_state_up'])
        self.assertEqual(status, port_details['status'])
        self.assertEqual(device_id, port_details['device_id'])
        self.assertEqual(device_owner, port_details['device_owner'])

    def _wait_for_port_detach(self, port_id, timeout=120, interval=10):
        """Waits for the port's device_id to be unset.

        :param port_id: The id of the port being detached.
        :returns: The final port dict from the show_port response.
        """
        port = self.client.show_port(port_id)['port']
        device_id = port['device_id']
        start = int(time.time())

        # NOTE(mriedem): Nova updates the port's device_id to '' rather than
        # None, but it's not contractual so handle Falsey either way.
        while device_id:
            time.sleep(interval)
            port = self.client.show_port(port_id)['port']
            device_id = port['device_id']

            timed_out = int(time.time()) - start >= timeout

            if device_id and timed_out:
                message = ('Port %s failed to detach (device_id %s) within '
                           'the required time (%s s).' %
                           (port_id, device_id, timeout))
                raise exceptions.TimeoutException(message)

        return port

    def _wait_for_fip_port_down(self, fip_id, timeout=120, interval=10):
        """Waits for the fip's attached port status to be 'DOWN'.

        :param fip_id: The id of the floating IP.
        :returns: The final fip dict from the show_floatingip response.
        """
        fip = self.client.show_floatingip(fip_id)['floatingip']
        self.assertIn('port_details', fip)
        port_details = fip['port_details']
        status = port_details['status']
        start = int(time.time())

        while status != lib_constants.PORT_STATUS_DOWN:
            time.sleep(interval)
            fip = self.client.show_floatingip(fip_id)['floatingip']
            self.assertIn('port_details', fip)
            port_details = fip['port_details']
            status = port_details['status']

            timed_out = int(time.time()) - start >= timeout

            if status != lib_constants.PORT_STATUS_DOWN and timed_out:
                port_id = fip.get("port_id")
                port = self.os_admin.network_client.show_port(port_id)['port']
                message = ('Floating IP %s attached port status failed to '
                           'transition to DOWN (current status %s) within '
                           'the required time (%s s). Port details: %s' %
                           (fip_id, status, timeout, port))
                raise exceptions.TimeoutException(message)

        LOG.debug('Port %s attached to FIP %s is down after %s!',
                  fip.get("port_id"), fip_id, int(time.time()) - start)
        return fip


class FloatingIPQosTest(FloatingIpTestCasesMixin,
                        test_qos.QoSTestMixin,
                        base.BaseTempestTestCase):

    same_network = True

    @classmethod
    @utils.requires_ext(extension="router", service="network")
    @utils.requires_ext(extension="qos", service="network")
    @utils.requires_ext(extension="qos-fip", service="network")
    @base_api.require_qos_rule_type(qos_consts.RULE_TYPE_BANDWIDTH_LIMIT)
    def resource_setup(cls):
        super().resource_setup()

    @classmethod
    def setup_clients(cls):
        super().setup_clients()
        cls.admin_client = cls.os_admin.network_client
        cls.qos_bw_limit_rule_client = \
            cls.os_admin.qos_limit_bandwidth_rules_client

    @decorators.idempotent_id('5eb48aea-eaba-4c20-8a6f-7740070a0aa3')
    def test_qos(self):
        """Test floating IP is binding to a QoS policy with

           ingress and egress bandwidth limit rules. And it applied correctly
           by sending a file from the instance to the test node.
           Then calculating the bandwidth every ~1 sec by the number of bits
           received / elapsed time.
        """

        self.skip_if_no_extension_enabled_in_l3_agents("fip_qos")

        self._test_basic_resources()

        # Create a new QoS policy
        policy_id = self._create_qos_policy()
        ssh_client = self._create_ssh_client()

        # As admin user create a new QoS rules
        rule_data = {'max_kbps': constants.LIMIT_KILO_BITS_PER_SECOND,
                     'max_burst_kbps': constants.LIMIT_KILO_BYTES,
                     'direction': lib_constants.INGRESS_DIRECTION}
        self.qos_bw_limit_rule_client.create_limit_bandwidth_rule(
             qos_policy_id=policy_id, **rule_data)

        rule_data = {'max_kbps': constants.LIMIT_KILO_BITS_PER_SECOND,
                     'max_burst_kbps': constants.LIMIT_KILO_BYTES,
                     'direction': lib_constants.EGRESS_DIRECTION}
        self.qos_bw_limit_rule_client.create_limit_bandwidth_rule(
             qos_policy_id=policy_id, **rule_data)

        rules = self.qos_bw_limit_rule_client.list_limit_bandwidth_rules(
            policy_id)
        self.assertEqual(2, len(rules['bandwidth_limit_rules']))

        fip = self.os_admin.network_client.get_floatingip(
            self.fip['id'])['floatingip']
        self.assertEqual(self.port['id'], fip['port_id'])

        # Associate QoS to the FIP
        self.os_admin.network_client.update_floatingip(
            self.fip['id'],
            qos_policy_id=policy_id)

        fip = self.os_admin.network_client.get_floatingip(
            self.fip['id'])['floatingip']
        self.assertEqual(policy_id, fip['qos_policy_id'])

        # Basic test, Check that actual BW while downloading file
        # is as expected (Original BW)
        common_utils.wait_until_true(lambda: self._check_bw(
            ssh_client,
            self.fip['floating_ip_address'],
            port=self.NC_PORT),
            timeout=120,
            sleep=1,
            exception=RuntimeError(
                'Failed scenario: "Create a QoS policy associated with FIP" '
                'Actual BW is not as expected!'))

        # As admin user update QoS rules
        for rule in rules['bandwidth_limit_rules']:
            self.qos_bw_limit_rule_client.update_limit_bandwidth_rule(
                policy_id, rule['id'],
                **{'max_kbps': constants.LIMIT_KILO_BITS_PER_SECOND * 2,
                   'max_burst_kbps': constants.LIMIT_KILO_BITS_PER_SECOND * 2})

        # Check that actual BW while downloading file
        # is as expected (Update BW)
        common_utils.wait_until_true(lambda: self._check_bw(
            ssh_client,
            self.fip['floating_ip_address'],
            port=self.NC_PORT,
            expected_bw=test_qos.QoSTestMixin.LIMIT_BYTES_SEC * 2),
            timeout=120,
            sleep=1,
            exception=RuntimeError(
                'Failed scenario: "Update QoS policy associated with FIP" '
                'Actual BW is not as expected!'))


class TestFloatingIPUpdate(FloatingIpTestCasesMixin,
                           base.BaseTempestTestCase):

    same_network = None

    @decorators.idempotent_id('1bdd849b-03dd-4b8f-994f-457cf8a36f93')
    def test_floating_ip_update(self):
        """Test updating FIP with another port.

        The test creates two servers and attaches floating ip to first server.
        Then it checks server is accesible using the FIP. FIP is then
        associated with the second server and connectivity is checked again.
        """
        ports = [self.create_port(
            self.network, security_groups=[self.secgroup['id']])
            for i in range(2)]

        servers = []
        for port in ports:
            name = data_utils.rand_name("server-%s" % port['id'][:8])
            server = self.create_server(
                name=name,
                flavor_ref=CONF.compute.flavor_ref,
                key_name=self.keypair['name'],
                image_ref=CONF.compute.image_ref,
                networks=[{'port': port['id']}])['server']
            server['name'] = name
            servers.append(server)
        for server in servers:
            self.wait_for_server_active(server)
            self.wait_for_guest_os_ready(server)

        self.fip = self.create_floatingip(port=ports[0])
        self.check_connectivity(self.fip['floating_ip_address'],
                                CONF.validation.image_ssh_user,
                                self.keypair['private_key'],
                                servers=servers)
        self.client.update_floatingip(self.fip['id'], port_id=ports[1]['id'])

        def _wait_for_fip_associated():
            try:
                self.check_servers_hostnames(servers[-1:], log_errors=False)
            except (AssertionError, exceptions.SSHTimeout):
                return False
            return True

        # The FIP is now associated with the port of the second server.
        try:
            common_utils.wait_until_true(_wait_for_fip_associated, sleep=3)
        except common_utils.WaitTimeout:
            self._log_console_output(servers[-1:])
            self.fail(
                "Server %s is not accessible via its floating ip %s" % (
                    servers[-1]['id'], self.fip['id']))


class FloatingIpMultipleRoutersTest(base.BaseTempestTestCase):
    credentials = ['primary', 'admin']

    @classmethod
    @utils.requires_ext(extension="router", service="network")
    def skip_checks(cls):
        super().skip_checks()

    def _create_keypair_and_secgroup(self):
        self.keypair = self.create_keypair()
        self.secgroup = self.create_security_group()
        self.create_loginable_secgroup_rule(
            secgroup_id=self.secgroup['id'])
        self.create_pingable_secgroup_rule(
            secgroup_id=self.secgroup['id'])

    def _delete_floating_ip(self, fip_address):
        ip_address = fip_address['floating_ip_address']

        def _fip_is_free():
            fips = self.os_admin.network_client.list_floatingips()
            for fip in fips['floatingips']:
                if ip_address == fip['floating_ip_address']:
                    return False
            return True

        self.delete_floatingip(fip_address)
        try:
            common_utils.wait_until_true(_fip_is_free, timeout=30, sleep=5)
        except common_utils.WaitTimeout:
            self.fail("Can't reuse IP address %s because it is not free" %
                      ip_address)

    def _create_network_and_servers(self, servers_num=1, fip_addresses=None,
                                    delete_fip_ids=None):
        delete_fip_ids = delete_fip_ids or []
        if fip_addresses:
            self.assertEqual(servers_num, len(fip_addresses),
                             ('Number of specified fip addresses '
                              'does not match the number of servers'))
        network = self.create_network()
        subnet = self.create_subnet(network)
        router = self.create_router_by_client()
        self.create_router_interface(router['id'], subnet['id'])

        fips = []
        for server in range(servers_num):
            fip = fip_addresses[server] if fip_addresses else None
            delete_fip = fip['id'] in delete_fip_ids if fip else False
            fips.append(
                self._create_server_and_fip(network=network,
                                            fip_address=fip,
                                            delete_fip_address=delete_fip))
        return fips

    def _create_server_and_fip(self, network, fip_address=None,
                               delete_fip_address=False):
        server = self.create_server(
            flavor_ref=CONF.compute.flavor_ref,
            image_ref=CONF.compute.image_ref,
            key_name=self.keypair['name'],
            networks=[{'uuid': network['id']}],
            security_groups=[{'name': self.secgroup['name']}])
        waiters.wait_for_server_status(self.os_primary.servers_client,
                                       server['server']['id'],
                                       constants.SERVER_STATUS_ACTIVE)
        port = self.client.list_ports(
            network_id=network['id'],
            device_id=server['server']['id'])['ports'][0]

        if fip_address:
            if delete_fip_address:
                self._delete_floating_ip(fip_address)
            fip = self.create_floatingip(
                floating_ip_address=fip_address['floating_ip_address'],
                client=self.os_admin.network_client,
                port=port)
            self.addCleanup(
                self.delete_floatingip, fip, self.os_admin.network_client)
        else:
            fip = self.create_floatingip(port=port)
        return fip

    def _check_fips_connectivity(self, mutable_fip, permanent_fip):
        for fip in [mutable_fip, permanent_fip]:
            fip['ssh_client'] = ssh.Client(fip['floating_ip_address'],
                                           CONF.validation.image_ssh_user,
                                           pkey=self.keypair['private_key'])
        self.check_remote_connectivity(
            permanent_fip['ssh_client'], mutable_fip['floating_ip_address'])
        self.check_remote_connectivity(
            mutable_fip['ssh_client'], permanent_fip['floating_ip_address'])

    @testtools.skipUnless(CONF.network.public_network_id,
                          'The public_network_id option must be specified.')
    @decorators.idempotent_id('b0382ab3-3c86-4415-84e3-649a8b040dab')
    def test_reuse_ip_address_with_other_fip_on_other_router(self):
        """Reuse IP address by another floating IP on another router

        Scenario:
            1. Create and connect a router to the external network.
            2. Create and connect an internal network to the router.
            3. Create and connect 2 VMs to the internal network.
            4. Create FIPs in the external network for the VMs.
            5. Make sure that VM1 can ping VM2 FIP address.
            6. Create and connect one more router to the external network.
            7. Create and connect an internal network to the second router.
            8. Create and connect a VM (VM3) to the internal network of
               the second router.
            9. Delete VM2 FIP but save IP address that it used. The FIP is
               deleted just before the creation of the new IP to "reserve" the
               IP address associated (see LP#1880976).
            10. Create a FIP for the VM3 in the external network with
                the same IP address that was used for VM2.
            11. Make sure that now VM1 is able to reach VM3 using the FIP.

        Note, the scenario passes only in case corresponding
        ARP update was sent to the external network when reusing same IP
        address for another FIP.
        """

        self._create_keypair_and_secgroup()
        [mutable_fip, permanent_fip] = (
            self._create_network_and_servers(servers_num=2))
        self._check_fips_connectivity(mutable_fip, permanent_fip)
        [mutable_fip] = self._create_network_and_servers(
            servers_num=1, fip_addresses=[mutable_fip],
            delete_fip_ids=[mutable_fip['id']])
        self._check_fips_connectivity(mutable_fip, permanent_fip)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils
from tempest.lib.common.utils import data_utils
from tempest.lib import decorators
from tempest.lib import exceptions as lib_exc

from neutron_tempest_plugin.api import base


class LoggingNegativeTestJSON(base.BaseAdminNetworkTest):

    required_extensions = ['logging', 'standard-attr-description']

    @decorators.attr(type='negative')
    @decorators.idempotent_id('5fc61e24-cad5-4d86-a2d4-f40c0fa0a54c')
    def test_create_log_with_invalid_resource_type(self):
        log_args = {'name': data_utils.rand_name('test-log'),
                    'description': data_utils.rand_name('test-log-desc'),
                    'resource_type': 'fake_resource'}
        self.assertRaises(lib_exc.BadRequest,
                          self.admin_client.create_log, **log_args)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('7ed63170-0748-44b7-b0a0-64bfd9390dac')
    def test_create_log_with_nonexistent_port(self):
        log_args = {'name': data_utils.rand_name('test-log'),
                    'description': data_utils.rand_name('test-log-desc'),
                    'resource_type': 'security_group',
                    'target_id': uuidutils.generate_uuid()}
        self.assertRaises(lib_exc.NotFound,
                          self.admin_client.create_log, **log_args)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('89194c6b-8f47-400b-979b-072b1c1f767b')
    def test_create_log_with_nonexistent_sg(self):
        log_args = {'name': data_utils.rand_name('test-log'),
                    'description': data_utils.rand_name('test-log-desc'),
                    'resource_type': 'security_group',
                    'resource_id': uuidutils.generate_uuid()}
        self.assertRaises(lib_exc.NotFound,
                          self.admin_client.create_log, **log_args)



#   Copyright 2021 Huawei, Inc. All rights reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from tempest.lib import decorators
from tempest.lib import exceptions

from neutron_tempest_plugin.api import base
from neutron_tempest_plugin import config

CONF = config.CONF


class LocalIPTestJSON(base.BaseNetworkTest):

    credentials = ['primary', 'admin']
    required_extensions = ['local_ip']

    @classmethod
    def resource_setup(cls):
        super().resource_setup()
        cls.ext_net_id = CONF.network.public_network_id

        # Create network and subnet
        cls.network = cls.create_network()
        cls.subnet = cls.create_subnet(cls.network)

    @decorators.idempotent_id('369257b0-521d-43f5-9482-50e18e87a472')
    def test_local_ip_lifecycle(self):
        port = self.create_port(self.network)
        lip_description = 'Test Local IP description'
        lip_name = 'test-local-ip'
        created_local_ip = self.create_local_ip(
            name=lip_name,
            description=lip_description,
            local_port_id=port['id'],
            local_ip_address=port['fixed_ips'][0]['ip_address'])
        self.assertEqual(self.network['id'], created_local_ip['network_id'])
        self.assertEqual(lip_description, created_local_ip['description'])
        self.assertEqual(lip_name, created_local_ip['name'])
        self.assertEqual(port['id'], created_local_ip['local_port_id'])
        self.assertEqual(port['fixed_ips'][0]['ip_address'],
                         created_local_ip['local_ip_address'])

        # Show created local_ip
        body = self.client.get_local_ip(created_local_ip['id'])
        local_ip = body['local_ip']

        self.assertEqual(lip_description, local_ip['description'])
        self.assertEqual(lip_name, local_ip['name'])

        # List local_ips
        body = self.client.list_local_ips()

        local_ip_ids = [lip['id'] for lip in body['local_ips']]
        self.assertIn(created_local_ip['id'], local_ip_ids)

        # Update local_ip
        updated_local_ip = self.client.update_local_ip(
                               created_local_ip['id'],
                               name='updated_local_ip')
        self.assertEqual('updated_local_ip',
                         updated_local_ip['local_ip']['name'])

        self.delete_local_ip(created_local_ip)
        self.assertRaises(exceptions.NotFound,
                          self.client.get_local_ip, created_local_ip['id'])

    @decorators.idempotent_id('e32df8ac-4e29-4adf-8057-46ae8684eff2')
    def test_create_local_ip_with_network(self):
        local_ip = self.create_local_ip(self.network['id'])
        self.assertEqual(self.network['id'], local_ip['network_id'])


class LocalIPAssociationTestJSON(base.BaseNetworkTest):

    required_extensions = ['local_ip']

    @classmethod
    def resource_setup(cls):
        super().resource_setup()
        cls.ext_net_id = CONF.network.public_network_id
        # Create network
        cls.network = cls.create_network()
        cls.subnet = cls.create_subnet(cls.network)

    @decorators.idempotent_id('602d2874-49be-4c72-8799-b20c95853b6b')
    def test_local_ip_association_lifecycle(self):
        local_ip = self.create_local_ip(self.network['id'])
        port = self.create_port(self.network)
        local_ip_association = self.create_local_ip_association(
            local_ip['id'],
            fixed_port_id=port['id'])
        self.assertEqual(local_ip['id'], local_ip_association['local_ip_id'])
        self.assertEqual(port['id'], local_ip_association['fixed_port_id'])

        # Test List Local IP Associations
        body = self.client.list_local_ip_associations(local_ip['id'])
        associations = body['port_associations']
        self.assertEqual(local_ip['id'], associations[0]['local_ip_id'])
        self.assertEqual(port['id'], associations[0]['fixed_port_id'])

        # Show
        body = self.client.get_local_ip_association(
            local_ip['id'], port['id'])
        association = body['port_association']
        self.assertEqual(local_ip['id'], association['local_ip_id'])
        self.assertEqual(port['id'], association['fixed_port_id'])

        # Delete
        self.client.delete_local_ip_association(local_ip['id'], port['id'])
        self.assertRaises(exceptions.NotFound,
                          self.client.get_local_ip_association,
                          local_ip['id'], port['id'])

    @decorators.idempotent_id('5d26edab-78d2-4cbd-9d0b-3c0b19f0f52d')
    def test_local_ip_association_with_two_ips_on_port(self):
        local_ip = self.create_local_ip(self.network['id'])
        s = self.subnet
        port = self.create_port(self.network)
        # request another IP on the same subnet
        port['fixed_ips'].append({'subnet_id': s['id']})
        updated = self.client.update_port(port['id'],
                                          fixed_ips=port['fixed_ips'])
        port = updated['port']
        local_ip_association = self.create_local_ip_association(
            local_ip['id'],
            fixed_port_id=port['id'],
            fixed_ip_address=port['fixed_ips'][0]['ip_address'])
        self.assertEqual(port['fixed_ips'][0]['ip_address'],
                         local_ip_association['fixed_ip'])
